/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.DBConnectionException;
import com.topcoder.db.connectionfactory.DBConnectionFactory;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.DocumentIndexerPersistenceHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.lang.reflect.InvocationTargetException;

import java.sql.Connection;
import java.sql.SQLException;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;


/**
 * <p>
 * An abstract <code>AbstractDBIndexPersistence</code> class which implements the <code>IndexPersistence</code>
 * interface.
 * </p>
 *
 * <p>
 * This class contains a connection name and a connectionFactory which will be used to get connection from DBConnection
 * factory in normal connection-obtaining schema. This class also provides the methods of transaction control.
 * </p>
 *
 * <p>
 * Thread-safe: this class is not thread safe.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public abstract class AbstractDBIndexPersistence implements IndexPersistence {
    /**
     * <p>
     * Represents the default class name of the connection factory. If the property value of connectionFactoryClassName
     * is missing in the config file, this default class name will be used to create the new connection factory
     * instance.
     * </p>
     */
    private static final String DEFAULT_CONNECTION_FACTORY_CLASS_NAME =
        "com.topcoder.db.connectionfactory.DBConnectionFactoryImpl";

    /**
     * <p>
     * Represents the connectionName property to get the connection from DBConnectionFactory. It is optional. If the
     * value of this property is missing, the default connection will be fetched.
     * </p>
     */
    private static final String CONNECTION_NAME_PROPERTY = "ConnectionName";

    /**
     * <p>
     * Represents class name property of connection factory to create the <code>DBConnectionFactory</code> instance via
     * reflection. It is optional. If the value of this property is missing, the default connectoin factory class name
     * will be used.
     * </p>
     */
    private static final String CLASS_NAME_PROPERTY = "ConnectionFactoryClassName";

    /**
     * <p>
     * Represents namespace property of connection factory to create the <code>DBConnectionFactory</code> instance via
     * reflection. It is required.
     * </p>
     */
    private static final String NAMESPACE_PROPERTY = "ConnectionFactoryNamespace";

    /**
     * <p>
     * Represents the connection factory from which we can obtain a pre-configured connection for our data base access.
     * </p>
     */
    private final DBConnectionFactory connectionFactory;

    /**
     * <p>
     * Represents a connection name (an alias) that is used to fetch a connection instance from the connection factory.
     * If the value of this field is <code>null</code> or empty, the default connection will be fetched.
     * </p>
     */
    private final String connectionName;

    /**
     * <p>
     * An abstract constructor of <code>AbstractDBIndexPersistence</code> class with given namespace.
     * </p>
     *
     * <p>
     * The property values of connectionName(optional), connectionFactoryClassName(optional), and
     * ConnectionFactoryNamespace(required) will be loaded from the configuration. A <code>DBConnectionFactory</code>
     * will be created via these property values. Connection will aslo be retrieved via this connection factory.
     * </p>
     *
     * @param namespace the namespace to load the property values required by this class.
     *
     * @throws IllegalArgumentException if <code>namespace</code> is <code>null</code> or empty string.
     * @throws PersistenceConfigurationException if it fails to load the property values required by this class or some
     *         of the property values is invalid.
     */
    protected AbstractDBIndexPersistence(String namespace) throws PersistenceConfigurationException {
        DocumentIndexerPersistenceHelper.validateString(namespace, "namespace");

        // get the connectionName property, this is optional
        this.connectionName = DocumentIndexerPersistenceHelper.getStringProperty(namespace, CONNECTION_NAME_PROPERTY,
                false);

        // get the ConnectionFactoryClassName property, this is optional
        String connectionFactoryClassName = DocumentIndexerPersistenceHelper.getStringProperty(namespace,
                CLASS_NAME_PROPERTY, false);

        // get the ConnectionFactoryNamespace property, this is required
        String connectionFactoryNamespace = DocumentIndexerPersistenceHelper.getStringProperty(namespace,
                NAMESPACE_PROPERTY, true);

        if (connectionFactoryClassName == null) {
            // if ConnectionFactoryClassName property does not exist, DBConnectionFactoryImpl will be used instead
            connectionFactoryClassName = DEFAULT_CONNECTION_FACTORY_CLASS_NAME;
        }

        try {
            // create the DBConnectionFactory instance with the given classname and namespace
            Class[] argClasses = new Class[] {String.class};
            Object[] argValues = new Object[] {connectionFactoryNamespace};
            this.connectionFactory = (DBConnectionFactory) Class.forName(connectionFactoryClassName)
                                                                .getConstructor(argClasses).newInstance(argValues);
        } catch (SecurityException e) {
            throw new PersistenceConfigurationException("Not enough privilege to access constructor.", e);
        } catch (InstantiationException e) {
            throw new PersistenceConfigurationException("The class is an abstract class.", e);
        } catch (IllegalAccessException e) {
            throw new PersistenceConfigurationException("The constructor cannot be accessed.", e);
        } catch (InvocationTargetException e) {
            throw new PersistenceConfigurationException("The constructor throws an exception.", e);
        } catch (NoSuchMethodException e) {
            throw new PersistenceConfigurationException("Proper constructor cannot be found.", e);
        } catch (ClassNotFoundException e) {
            throw new PersistenceConfigurationException("The word source class cannot be found.", e);
        } catch (ClassCastException e) {
            throw new PersistenceConfigurationException("The class is not a DBConnectionFactory class.", e);
        }
    }

    /**
     * <p>
     * An abstract constructor of <code>AbstractDBIndexPersistence</code> class with given connectionFactory and
     * connectionName information from input parameters.
     * </p>
     *
     * @param connectionFactory the connection factory from which we can obtain a pre-configured connection for our
     *        data base access.
     * @param connectionName a connection name (an alias) that is used to fetch a connection instance from the
     *        connection factory.
     *
     * @throws IllegalArgumentException if <code>connectionFactory</code> is <code>null</code>
     */
    protected AbstractDBIndexPersistence(DBConnectionFactory connectionFactory, String connectionName) {
        DocumentIndexerPersistenceHelper.validateNotNull(connectionFactory, "connectionFactory");
        this.connectionFactory = connectionFactory;
        this.connectionName = connectionName;
    }

    /**
     * <p>
     * Gets a new a connection to be used for persistence. The auto-commit value of the connection returned is set to
     * <code>false</code> in this method.
     * </p>
     *
     * @return a new connection to be used for persistence.
     *
     * @throws IndexPersistenceException If unable to obtain a connection or set the value of auto-commit.
     */
    protected Connection getConnection() throws IndexPersistenceException {
        try {
            Connection connection = null;

            if ((connectionName == null) || (connectionName.trim().length() == 0)) {
                // the default connection name will be used.
                connection = this.connectionFactory.createConnection();
            } else {
                connection = this.connectionFactory.createConnection(connectionName);
            }

            // set the value of auto-commit to be false
            connection.setAutoCommit(false);

            return connection;
        } catch (SQLException e) {
            throw new IndexPersistenceException("Unable to set the value of auto-commit.", e);
        } catch (DBConnectionException e) {
            throw new IndexPersistenceException("Unable to obtain a connection.", e);
        }
    }

    /**
     * <p>
     * Starts a transaction on the passed in connection. Do nothing here since auto commit is already set to
     * <code>false</code>.
     * </p>
     *
     * @param connection the connection who takes part in the transaction.
     *
     * @throws IllegalArgumentException if <code>connection</code> is <code>null</code>
     * @throws IndexPersistenceException If an error occurs during this process, or if the connection does not support
     *         transactions.
     */
    protected void beginTransaction(Connection connection) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(connection, "connection");

        // do nothing here since auto commit is already false
    }

    /**
     * <p>
     * Commits a transaction on the passed in connection.
     * </p>
     *
     * @param connection the connection who takes part in the transaction.
     *
     * @throws IllegalArgumentException if <code>connection</code> is <code>null</code>
     * @throws IndexPersistenceException If an error occurs during this process, or if the connection does not support
     *         transactions.
     */
    protected void commitTransaction(Connection connection) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(connection, "connection");

        try {
            connection.commit();
            connection.close();
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error happens during the commit process.", e);
        }
    }

    /**
     * <p>
     * Rolls back the transaction on the passed in connection.
     * </p>
     *
     * @param connection the connection who takes part in the transaction.
     *
     * @throws IllegalArgumentException if <code>connection</code> is <code>null</code>
     * @throws IndexPersistenceException If an error occurs during this process, or if the connection does not support
     *         transactions.
     */
    protected void rollbackTransaction(Connection connection) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(connection, "connection");

        try {
            connection.rollback();
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error happens during the rollback process.", e);
        } finally {
            try {
                connection.close();
            } catch (SQLException e) {
                // ignore it
            }
        }
    }

    /**
     * <p>
     * Adds the <code>WordSourceId</code> into the database persistence.
     * </p>
     *
     * @param connection the connection to access database.
     * @param wordSourceId the <code>WordSourceId</code> to add into the database persistence.
     * @param documentId the document id of the <code>WordSourceId</code> to add into the database persistence.
     *
     * @throws IndexPersistenceException if if fails to add the wordSourceId into the database persistence.
     */
    void addWordSourceId(Connection connection, WordSourceId wordSourceId, String documentId)
        throws IndexPersistenceException {
        try {
            String locale = wordSourceId.getSourceLocale().toString();
            byte[] sourceIdentity = DocumentIndexerPersistenceHelper.getSerializedBytes(
                    wordSourceId.getSourceIdentity());
            String className = wordSourceId.getSourceClassName();
            SQLUpdates.insertDocumentTable(connection, documentId, locale, sourceIdentity, className, 0);

            // 2: Create a DELIMITER record for each member of the delimiters array in the wordSourceId variable.
            for (int i = 0; i < wordSourceId.getDelimiters().length; i++) {
                String delimiter = wordSourceId.getDelimiters()[i];
                SQLUpdates.insertDelimiterTable(connection, documentId, delimiter, i);
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to add the wordSourceId into the database persistence.");
        }
    }

    /**
     * <p>
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails.
     * </p>
     *
     * @param connection the connection to access database.
     * @param collectionIndex the collection index to be persisted.
     *
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in
     *         the persistence.
     */
    void addCollectionIndex(Connection connection, CollectionIndex collectionIndex) throws IndexPersistenceException {
        try {
            String collectionId = collectionIndex.getId();

            if (SQLQueries.containsDocumentCollection(connection, collectionId)) {
                throw new IndexPersistenceException("The collection index exists in the persistence.");
            }

            // 1: create a new DOCUMENT_COLLECTION record
            SQLUpdates.insertDocumentCollectionTable(connection, collectionId);

            // 2. update the records in the the DOCUMENT and DOCUMENT_COLLECTION_DOCUMENT_XREF TABLE
            for (Iterator iter = collectionIndex.getAllDocumentIds().iterator(); iter.hasNext();) {
                String documentId = DocumentIndexerPersistenceHelper.getDocumentId((WordSourceId) iter.next());

                // create a new DOCUMENT_COLLECTION_DOCUMENT_XREF record
                SQLUpdates.insertDocumentCollectionDocumentXref(connection, collectionId, documentId);
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error in the adding process.", e);
        }
    }

    /**
     * <p>
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails.
     * </p>
     *
     * @param connection the connection to access database.
     * @param collectionId the ID of the collection index to be deleted.
     *
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    void removeCollectionIndex(Connection connection, String collectionId) throws IndexPersistenceException {
        try {
            if (!SQLQueries.containsDocumentCollection(connection, collectionId)) {
                throw new IndexPersistenceException("The collection index ID cannot be found in the persistence.");
            }

            // 1: delete the records in DOCUMENT_COLLECTION_DOCUMENT_XREF table
            SQLUpdates.deleteDocumentCollectionDocumentXrefTable(connection, collectionId);

            // 2: delete the record in DOCUMENT_COLLECTION table
            SQLUpdates.deleteDocumentCollectionTable(connection, collectionId);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error in the removing process.", e);
        }
    }

    /**
     * <p>
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the
     * persistence, it fails.
     * </p>
     *
     * @param connection the connection to access database.
     * @param collectionIndex the collection index to be updated.
     *
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    void updateCollectionIndex(Connection connection, CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        try {
            String collectionId = collectionIndex.getId();

            if (!SQLQueries.containsDocumentCollection(connection, collectionId)) {
                throw new IndexPersistenceException("The collection index ID cannot be found in the persistence.");
            }

            // if there is any document index entries there that are not in the input CollectionIndex object,
            // remove the records in the DOCUMENT_COLLECTION_DOCUMENT_XREF table
            Set documentIdsToDelete = new HashSet(SQLQueries.getDocumentIds(connection, collectionId));

            // if there any document index entries in the CollectionIndex object that are not in the current
            // persistence, add the records to the DOCUMENT_COLLECTION_DOCUMENT_XREF table
            List documentIdsToAdd = new ArrayList();

            for (Iterator iter = collectionIndex.getAllDocumentIds().iterator(); iter.hasNext();) {
                String documentId = DocumentIndexerPersistenceHelper.getDocumentId((WordSourceId) iter.next());

                if (!documentIdsToDelete.contains(documentId)) {
                    documentIdsToAdd.add(documentId);
                } else {
                    documentIdsToDelete.remove(documentId);
                }
            }

            // remove the records in the DOCUMENT_COLLECTION_DOCUMENT_XREF table
            for (Iterator iter = documentIdsToDelete.iterator(); iter.hasNext();) {
                String documentId = (String) iter.next();
                SQLUpdates.deleteDocumentCollectionDocumentXrefTable(connection, collectionId, documentId);
            }

            //  add the records to the DOCUMENT_COLLECTION_DOCUMENT_XREF table
            for (Iterator iter = documentIdsToAdd.iterator(); iter.hasNext();) {
                String documentId = (String) iter.next();
                SQLUpdates.insertDocumentCollectionDocumentXref(connection, collectionId, documentId);
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error in the updating process.", e);
        }
    }

    /**
     * <p>
     * Increases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * @param connection the connection to access database.
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     *
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    void increaseDocumentUseCount(Connection connection, WordSourceId wordSourceId) throws IndexPersistenceException {
        try {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);

            if (!SQLQueries.containsDocument(connection, documentId)) {
                // no document with given id exist
                throw new IndexPersistenceException("The document index is not persisted.");
            }

            // update the use count
            int useCount = SQLQueries.getDocumentUseCount(connection, documentId);
            SQLUpdates.updateDocumentUseCount(connection, documentId, useCount + 1);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error in the increasing process.", e);
        }
    }

    /**
     * <p>
     * Decreases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * @param connection the connection to access database.
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     *
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *         not persisted; or the collection reference count is already 0.
     */
    void decreaseDocumentUseCount(Connection connection, WordSourceId wordSourceId) throws IndexPersistenceException {
        try {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);

            if (!SQLQueries.containsDocument(connection, documentId)) {
                // no document with given id exist
                throw new IndexPersistenceException("The document index is not persisted.");
            }

            // update the use count
            int useCount = SQLQueries.getDocumentUseCount(connection, documentId);

            if (useCount == 0) {
                throw new IndexPersistenceException("The collection reference count is already 0.");
            }

            SQLUpdates.updateDocumentUseCount(connection, documentId, useCount - 1);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error in the decreasing process.", e);
        }
    }
}
