/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.DBConnectionFactory;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.DocumentIndexerPersistenceHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.sql.Connection;
import java.sql.SQLException;

import java.text.CollationKey;
import java.text.Collator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * <p>
 * Implements the <code>AbstractDBIndexPersistence</code> which deals with a fast, CLOB based, normalized data based
 * schema. This is provided for a user how is concerned about speed of the execution. By storing all the word and word
 * position data in a single CLOB structure it can achieve a 4000% speed increase over a comparable solution that uses
 * thousands of records to store words and related positions.
 * </p>
 *
 * <p>
 * Thread-safe: this class is not thread safe.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public class FastAccessDBIndexPersistence extends AbstractDBIndexPersistence {
    /**
     * <p>
     * Creates a new instance of <code>FastAccessDBIndexPersistence</code> class with given namespace.
     * </p>
     *
     * <p>
     * The property values of connectionName(optional), connectionFactoryClassName(optional), and
     * ConnectionFactoryNamespace(required) will be loaded from the configuration. A <code>DBConnectionFactory</code>
     * will be created via these property values. Connection will aslo be retrieved via this connection factory.
     * </p>
     *
     * @param namespace the namespace to load the property values required by this class.
     *
     * @throws IllegalArgumentException if <code>namespace</code> is <code>null</code> or empty string.
     * @throws PersistenceConfigurationException if it fails to load the property values required by this class or some
     *         of the property values is invalid.
     */
    public FastAccessDBIndexPersistence(String namespace) throws PersistenceConfigurationException {
        super(namespace);
    }

    /**
     * <p>
     * Creates a new instance of <code>FastAccessDBIndexPersistence</code> class with given connectionFactory and
     * connectionName information from input parameters.
     * </p>
     *
     * @param connectionFactory the connection factory from which we can obtain a pre-configured connection for our
     *        data base access.
     * @param connectionName a connection name (an alias) that is used to fetch a connection instance from the
     *        connection factory.
     *
     * @throws IllegalArgumentException if <code>connectionFactory</code> is <code>null</code>
     */
    public FastAccessDBIndexPersistence(DBConnectionFactory connectionFactory, String connectionName) {
        super(connectionFactory, connectionName);
    }

    /**
     * <p>
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails.
     * </p>
     *
     * @param documentIndex the document index to be persisted.
     *
     * @throws IllegalArgumentException if <code>documentIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *         persistence.
     */
    public void addDocumentIndex(DocumentIndex documentIndex) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(documentIndex, "documentIndex");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId(documentIndex.getWordSourceId());

            // check whether the document index exists in the persistence.
            if (SQLQueries.containsDocument(connection, documentId)) {
                throw new IndexPersistenceException("The document index exists in the persistence.");
            }

            // 1: add the wordSourceId record.
            super.addWordSourceId(connection, documentIndex.getWordSourceId(), documentId);

            // 2: create a CLOB word data for each entry in the words Map
            String posDelimiter = documentIndex.getWordSourceId().getDelimiters()[0];
            String wordDelimiter = posDelimiter + posDelimiter;

            StringBuffer buffer = new StringBuffer();
            boolean first = true;

            for (Iterator iter = documentIndex.getWords().entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();

                // get the word
                String word = ((CollationKey) entry.getKey()).getSourceString();

                if (!first) {
                    buffer.append(wordDelimiter);
                }

                first = false;

                buffer.append(word);

                // get the positions of the word
                List positions = (List) entry.getValue();

                // for each position, create a new WORD_POSITION record
                for (Iterator positionIter = positions.iterator(); positionIter.hasNext();) {
                    buffer.append(posDelimiter).append(positionIter.next());
                }
            }

            // 3:  store new WORD_DATA record.
            SQLUpdates.insertWordDataTable(connection, documentId, buffer.toString());

            // commit the transaction
            super.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw e;
        } catch (SQLException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw new IndexPersistenceException("Error in the adding process.", e);
        }
    }

    /**
     * <p>
     * Gets the document index from the persistence. The ID of the document is given. When the document is not
     * persisted in the persistence, <code>null</code> should be returned.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is retrieved.
     *
     * @return the document index of the document with the ID.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the document index fails.
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);

            if (!SQLQueries.containsDocument(connection, documentId)) {
                return null;
            }

            // initialize locale-sensitive comparator
            Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());
            collator.setStrength(Collator.PRIMARY);

            // get the WORD_DATA record.
            String wordData = SQLQueries.getWordData(connection, documentId);

            // fetch all the words and their positions
            String posDelimiter = wordSourceId.getDelimiters()[0];
            Map words = this.getWords(wordData, posDelimiter, collator);

            return new DocumentIndex(wordSourceId, words);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error happens in the getting document id process.", e);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * Parsing all the words and their positions from the CLOB record from the WORD_DATA table.
     *
     * @param wordData the CLOB record from the WORD_DATA table.
     * @param posDelimiter the position delimiter of the parsing process.
     * @param collator the collator from which the word will get the collatorKey.
     *
     * @return all the words and their positions.
     *
     * @throws IndexPersistenceException if parsing the CLOB record fails.
     */
    private Map getWords(String wordData, String posDelimiter, Collator collator) throws IndexPersistenceException {
        Map words = new HashMap();
        int pos = 0;
        boolean readWordState = true;
        boolean readPositionState = true;
        String word = null;
        List positions = null;

        wordData += posDelimiter;

        while (true) {
            int index = wordData.indexOf(posDelimiter, pos);

            if (index < 0) {
                index = wordData.length();
            }

            if (readWordState) {
                // in state of reading word
                word = wordData.substring(pos, index);
                positions = new ArrayList();
                readWordState = false;
                readPositionState = true;
            } else if (readPositionState) {
                // in state of reading positions
                positions.add(new Integer(this.parseInt(wordData.substring(pos, index))));
            }

            pos = index + posDelimiter.length();

            // means a new word begins
            index = wordData.indexOf(posDelimiter, pos);

            if ((index == -1) || (pos == index)) {
                pos += posDelimiter.length();
                readWordState = true;
                readPositionState = false;
            }

            if (readWordState) {
                words.put(collator.getCollationKey(word), positions);
            }

            if (pos >= wordData.length()) {
                break;
            }
        }

        return words;
    }

    /**
     * Parse the string value into int value.
     *
     * @param stringValue the string value to parse into the int value.
     *
     * @return the parsed int value.
     *
     * @throws IndexPersistenceException if the int value is not well formatted.
     */
    private int parseInt(String stringValue) throws IndexPersistenceException {
        try {
            return Integer.parseInt(stringValue);
        } catch (NumberFormatException e) {
            throw new IndexPersistenceException("The int value of " + stringValue + " is not well formatted", e);
        }
    }

    /**
     * <p>
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The ID of the document is given. When the document is not persisted,
     * it fails.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is deleted.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *         or there is at least one collection containing the document index.
     */
    public void removeDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);

            if (!SQLQueries.containsDocument(connection, documentId)) {
                throw new IndexPersistenceException("The document index is not persisted.");
            }

            if (SQLQueries.getDocumentUseCount(connection, documentId) != 0) {
                throw new IndexPersistenceException("There is at least one collection containing the document index.");
            }

            // Delete all the WORD_DATA records
            SQLUpdates.deleteWordDataTable(connection, documentId);

            // Delete all the DELIMITER records
            SQLUpdates.deleteDelimiterTable(connection, documentId);

            // No need to delete the DOCUMENT_COLLECTION_DOCUMENT_XREF records since there should be no record left
            // Delete the DOCUMENT record
            SQLUpdates.deleteDocumentTable(connection, documentId);

            // commit the transaction
            super.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw e;
        } catch (SQLException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw new IndexPersistenceException("Error in the removing process.", e);
        }
    }

    /**
     * <p>
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails.
     * </p>
     *
     * @param collectionIndex the collection index to be persisted.
     *
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in
     *         the persistence.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(collectionIndex, "collectionIndex");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            // add the collection index
            super.addCollectionIndex(connection, collectionIndex);

            // commit the transaction
            super.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw e;
        }
    }

    /**
     * <p>
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, <code>null</code> should be returned.
     * </p>
     *
     * @param collectionId the ID of the collection index to be retrieved.
     *
     * @return the collection index with the given ID.
     *
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    public CollectionIndex getCollectionIndex(String collectionId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateString(collectionId, "collectionId");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            if (!SQLQueries.containsDocumentCollection(connection, collectionId)) {
                return null;
            }

            List ids = SQLQueries.getDocumentIds(connection, collectionId);
            Map words = new HashMap();
            Set allDocumentIds = new HashSet();

            for (Iterator iter = ids.iterator(); iter.hasNext();) {
                String documentId = (String) iter.next();

                // get the wordSourceId
                WordSourceId wordSourceId = SQLQueries.getWordSourceId(connection, documentId);
                Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());
                collator.setStrength(Collator.PRIMARY);

                // get all the words
                String wordData = SQLQueries.getWordData(connection, documentId);
                Set wordsSet = this.getWords(wordData, wordSourceId.getDelimiters()[0], collator).keySet();

                // Add all words into the words map
                for (Iterator wordsSetIterator = wordsSet.iterator(); wordsSetIterator.hasNext();) {
                    CollationKey key = (CollationKey) wordsSetIterator.next();
                    Set documentIds = (Set) words.get(key);

                    if (documentIds == null) {
                        // The word does not exist, create a new list
                        documentIds = new HashSet();
                        words.put(key, documentIds);
                    }

                    // Mark the word's existence in the document
                    documentIds.add(wordSourceId);
                }

                // add the wordSourceId
                allDocumentIds.add(wordSourceId);
            }

            return new CollectionIndex(null, words, allDocumentIds, collectionId);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error happens in the getting collection id process.", e);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * <p>
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails.
     * </p>
     *
     * @param collectionId the ID of the collection index to be deleted.
     *
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    public void removeCollectionIndex(String collectionId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateString(collectionId, "collectionId");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            super.removeCollectionIndex(connection, collectionId);

            // commit the transaction
            super.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw e;
        }
    }

    /**
     * <p>
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the
     * persistence, it fails.
     * </p>
     *
     * @param collectionIndex the collection index to be updated.
     *
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(collectionIndex, "collectionIndex");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            super.updateCollectionIndex(connection, collectionIndex);

            // commit the transaction
            super.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw e;
        }
    }

    /**
     * <p>
     * Increases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            super.increaseDocumentUseCount(connection, wordSourceId);

            // commit the transaction
            super.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw e;
        }
    }

    /**
     * <p>
     * Decreases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *         not persisted; or the collection reference count is already 0.
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            super.decreaseDocumentUseCount(connection, wordSourceId);

            // commit the transaction
            super.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            // rollback the transaction if any errors happen
            super.rollbackTransaction(connection);
            throw e;
        }
    }

    /**
     * <p>
     * Gets a set of all IDs of documents whose indices are persisted in the persistence.
     * </p>
     *
     * @return a set of all IDs of persisted document indices.
     *
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        // get the connection.
        Connection connection = super.getConnection();

        try {
            return new HashSet(SQLQueries.getWordSourceIds(connection));
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error happens in the getting index documents process.", e);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * <p>
     * Gets the reference count of the given document index. When the document is not in the persistence, return 0.
     * </p>
     *
     * @param wordSourceId the ID of the document index whose reference count is retrieved.
     *
     * @return the reference count of the document index.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *         not persisted.
     */
    public int getDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        // get the connection.
        Connection connection = super.getConnection();

        try {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);

            return SQLQueries.getDocumentUseCount(connection, documentId);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error happens in the getting index documents process.", e);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    // ignore
                }
            }
        }
    }
}
