/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.DocumentIndexerPersistenceHelper;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;


/**
 * <p>
 * This class contains methods with SQL queries used by <code>FastAccessDBIndexPersistence</code> and
 * <code>SimpleAccessDBIndexPersistence</code>.
 * </p>
 *
 * <p>
 * Each method accepts a connection argument, and usually represents single sql statement.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
final class SQLQueries {
    /** Represents the column name for document id. */
    private static final String DOCUMENT_ID_COLUMN = "DOCUMENT_ID";

    /** Represents the column name for document locale. */
    private static final String LOCALE_COLUMN = "LOCALE";

    /** Represents the column name for document source identity. */
    private static final String SOURCE_IDENTITY_COLUMN = "SOURCE_IDENTITY";

    /** Represents the column name for document class name. */
    private static final String CLASS_NAME_COLUMN = "CLASS_NAME";

    /** Represents the column name for document use count. */
    private static final String USE_COUNT_COLUMN = "USE_COUNT";

    /** Represents the column name for delimiter entry. */
    private static final String DELIMITER_ENTRY_COLUMN = "DELIMITER_ENTRY";

    /** Represents the column name for word id. */
    private static final String WORD_ID_COLUMN = "WORD_ID";

    /** Represents the column name for word entry. */
    private static final String WORD_ENTRY_COLUMN = "WORD_ENTRY";

    /** Represents the column name for word position. */
    private static final String WORD_POSITION_COLUMN = "WORD_POSITION";

    /** Represents the column name for word data. */
    private static final String WORD_DATA_COLUMN = "WORD_DATA";

    /** Represents the prepared SQL statement to get all the records from DOCUMENT table with given DOCUMENT_ID. */
    private static final String RETRIEVE_DOCUMENT_SQL = "SELECT LOCALE, SOURCE_IDENTITY, CLASS_NAME FROM "
        + "DOCUMENT WHERE DOCUMENT_ID = ?";

    /** Represents the prepared SQL statement to get all the records from DOCUMENT table. */
    private static final String RETRIEVE_ALL_DOCUMENT_SQL = "SELECT DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, "
        + "CLASS_NAME FROM DOCUMENT";

    /** Represents the prepared SQL statement to get the use count record from DOCUMENT table with given DOCUMENT_ID. */
    private static final String RETRIEVE_USECOUNT_SQL = "SELECT USE_COUNT FROM DOCUMENT WHERE DOCUMENT_ID = ?";

    /** Represents the prepared SQL statement to get the records from DELIMITER table with given DOCUMENT_ID. */
    private static final String RETRIEVE_DELIMITER_SQL = "SELECT DELIMITER_ENTRY FROM DELIMITER "
            + "WHERE DOCUMENT_ID = ? ORDER BY DELIMITER_SORT_ORDER";

    /** Represents the prepared SQL statement to get the records from WORD table with given DOCUMENT_ID. */
    private static final String RETRIEVE_WORD_SQL = "SELECT WORD_ID, WORD_ENTRY FROM WORD WHERE DOCUMENT_ID = ?";

    /**
     * Represents the prepared SQL statement to get the records from WORD_POSITION table with given DOCUMENT_ID and
     * WORD_ID.
     */
    private static final String RETRIEVE_WORD_POSITION_SQL = "SELECT WORD_POSITION FROM WORD_POSITION WHERE "
        + "DOCUMENT_ID = ? AND WORD_ID = ?";

    /**
     * Represents the prepared SQL statement to get the records from DOCUMENT_COLLECTION table with given
     * COLLECTION_ID.
     */
    private static final String RETRIEVE_DOCUMENT_COLLECTION_SQL = "SELECT COLLECTION_ID FROM DOCUMENT_COLLECTION "
        + "WHERE COLLECTION_ID = ?";

    /**
     * Represents the prepared SQL statement to get the records from DOCUMENT_COLLECTION_DOCUMENT_XREF table with given
     * COLLECTION_ID.
     */
    private static final String RETRIEVE_DOCUMENT_COLLECTION_DOCUMENT_XREF_SQL = "SELECT DOCUMENT_ID FROM "
        + "DOCUMENT_COLLECTION_DOCUMENT_XREF WHERE COLLECTION_ID = ?";

    /** Represents the prepared SQL statement to get the records from WORD_DATA table with given DOCUMENT_ID. */
    private static final String RETRIEVE_WORD_DATA_SQL = "SELECT WORD_DATA FROM WORD_DATA WHERE DOCUMENT_ID = ?";

    /**
     * Private constructor to prevent this class be instantiated.
     */
    private SQLQueries() {
        // empty
    }

    /**
     * <p>
     * Retrieves a <code>WordSourceId</code> with given document id from the database records.
     * </p>
     *
     * @param connection the connection to access database.
     * @param documentId the given document id to get the <code>WordSourceId</code>.
     *
     * @return a <code>WordSourceId</code> with given document id from the database records.
     *
     * @throws SQLException if a database error occurs.
     * @throws IndexPersistenceException if the locale string is not well formatted or the sourceIdentity can not be
     *         created via the data from the database.
     */
    public static WordSourceId getWordSourceId(Connection connection, String documentId)
        throws SQLException, IndexPersistenceException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_DOCUMENT_SQL);
            statement.setString(1, documentId);
            resultSet = statement.executeQuery();

            if (!resultSet.next()) {
                return null;
            }

            String localeString = resultSet.getString(LOCALE_COLUMN);
            Locale locale = DocumentIndexerPersistenceHelper.parseLocaleString(localeString);
            byte[] sourceIdentityBytes = resultSet.getBytes(SOURCE_IDENTITY_COLUMN);
            Object sourceIdentity = DocumentIndexerPersistenceHelper.getSerializedObject(sourceIdentityBytes, 0,
                    sourceIdentityBytes.length);
            String className = resultSet.getString(CLASS_NAME_COLUMN);
            List delimiters = getDelimiters(connection, documentId);

            return new WordSourceId(sourceIdentity, className,
                (String[]) delimiters.toArray(new String[delimiters.size()]), locale);
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Retrieves the word data with given document id from the database records.
     * </p>
     *
     * @param connection the connection to access database.
     * @param documentId the given document id to get the word data.
     *
     * @return the word data with given document id from the database records.
     *
     * @throws SQLException if a database error occurs.
     */
    public static String getWordData(Connection connection, String documentId) throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_WORD_DATA_SQL);
            statement.setString(1, documentId);
            resultSet = statement.executeQuery();

            if (!resultSet.next()) {
                return null;
            }

            return resultSet.getString(WORD_DATA_COLUMN);
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Retrieves the list of word source ids from the database records.
     * </p>
     *
     * @param connection the connection to access database.
     *
     * @return the list of word source ids from the database records.
     *
     * @throws SQLException if a database error occurs.
     * @throws IndexPersistenceException if the locale string is not well formatted or the sourceIdentity can not be
     *         created via the data from the database.
     */
    public static List getWordSourceIds(Connection connection) throws SQLException, IndexPersistenceException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_ALL_DOCUMENT_SQL);
            resultSet = statement.executeQuery();

            List wordSourceIds = new ArrayList();

            while (resultSet.next()) {
                String documentId = resultSet.getString(DOCUMENT_ID_COLUMN);
                String localeString = resultSet.getString(LOCALE_COLUMN);
                Locale locale = DocumentIndexerPersistenceHelper.parseLocaleString(localeString);
                byte[] sourceIdentityBytes = resultSet.getBytes(SOURCE_IDENTITY_COLUMN);
                Object sourceIdentity = DocumentIndexerPersistenceHelper.getSerializedObject(sourceIdentityBytes, 0,
                        sourceIdentityBytes.length);
                String className = resultSet.getString(CLASS_NAME_COLUMN);
                List delimiters = getDelimiters(connection, documentId);

                wordSourceIds.add(new WordSourceId(sourceIdentity, className,
                        (String[]) delimiters.toArray(new String[delimiters.size()]), locale));
            }

            return wordSourceIds;
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Retrieves the list of delimiters from the database records with given document id.
     * </p>
     *
     * @param connection the connection to access database.
     * @param documentId the given document id to get the list of delimiters.
     *
     * @return the list of delimiters from the database records with given document id.
     *
     * @throws SQLException if a database error occurs.
     */
    private static List getDelimiters(Connection connection, String documentId) throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_DELIMITER_SQL);
            statement.setString(1, documentId);
            resultSet = statement.executeQuery();

            List delimiters = new ArrayList();

            while (resultSet.next()) {
                String delimiter = resultSet.getString(DELIMITER_ENTRY_COLUMN);
                StringBuffer buffer = new StringBuffer();
                // decoding it since we have escaped the character of '\0'
                int i = 0;
                for (i = 0; i < delimiter.length() - 1; i++) {
                    if (delimiter.charAt(i) == '\\' && delimiter.charAt(i + 1) == '\\') {
                        buffer.append('\\');
                        i++;
                    } else if (delimiter.charAt(i) == '\\' && delimiter.charAt(i + 1) == '0') {
                        buffer.append('\0');
                        i++;
                    } else {
                        buffer.append(delimiter.charAt(i));
                    }
                }
                if (i == delimiter.length() - 1) {
                    buffer.append(delimiter.charAt(i));
                }
                delimiters.add(buffer.toString());
            }

            return delimiters;
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Retrieves the list of words from the database records with given document id.
     * </p>
     *
     * @param connection the connection to access database.
     * @param documentId the given document id to get the list of words.
     *
     * @return the list of words from the database records with given document id.
     *
     * @throws SQLException if a database error occurs.
     */
    public static List getWords(Connection connection, String documentId) throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_WORD_SQL);
            statement.setString(1, documentId);
            resultSet = statement.executeQuery();

            List words = new ArrayList();

            while (resultSet.next()) {
                words.add(new Integer(resultSet.getInt(WORD_ID_COLUMN)));
                words.add(resultSet.getString(WORD_ENTRY_COLUMN));
            }

            return words;
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Retrieves the list of word positions from the database records with given document id and word id.
     * </p>
     *
     * @param connection the connection to access database.
     * @param documentId the given document id to get the list of word positions.
     * @param wordId the given word id to get the list of word positions.
     *
     * @return the list of word positions from the database records with given document id and word id.
     *
     * @throws SQLException if a database error occurs.
     */
    public static List getWordPositions(Connection connection, String documentId, int wordId)
        throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_WORD_POSITION_SQL);
            statement.setString(1, documentId);
            statement.setInt(2, wordId);
            resultSet = statement.executeQuery();

            List positions = new ArrayList();

            while (resultSet.next()) {
                positions.add(new Integer(resultSet.getInt(WORD_POSITION_COLUMN)));
            }

            return positions;
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Retrieves the list of document ids from the database records with given collection id.
     * </p>
     *
     * @param connection the connection to access database.
     * @param collectionId the given collection id to get the list of document ids.
     *
     * @return the list of document ids from the database records with given collection id.
     *
     * @throws SQLException if a database error occurs.
     */
    public static List getDocumentIds(Connection connection, String collectionId) throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_DOCUMENT_COLLECTION_DOCUMENT_XREF_SQL);
            statement.setString(1, collectionId);
            resultSet = statement.executeQuery();

            List documentIds = new ArrayList();

            while (resultSet.next()) {
                documentIds.add(resultSet.getString(DOCUMENT_ID_COLUMN));
            }

            return documentIds;
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Checks if there is a record with given the document id in the database.
     * </p>
     *
     * @param connection the connection to access database.
     * @param documentId the document id to check.
     *
     * @return whether there is a record with given the document id in the database.
     *
     * @throws SQLException if a database error occurs.
     */
    public static boolean containsDocument(Connection connection, String documentId) throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_USECOUNT_SQL);
            statement.setString(1, documentId);
            resultSet = statement.executeQuery();

            return resultSet.next();
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Checks if there is a record with given the collection id in the database.
     * </p>
     *
     * @param connection the connection to access database.
     * @param collectionId the collection id to check.
     *
     * @return whether there is a record with given the collection id in the database.
     *
     * @throws SQLException if a database error occurs.
     */
    public static boolean containsDocumentCollection(Connection connection, String collectionId)
        throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_DOCUMENT_COLLECTION_SQL);
            statement.setString(1, collectionId);
            resultSet = statement.executeQuery();

            return resultSet.next();
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Retrieves the use count from the database records with given document id.
     * </p>
     *
     * @param connection the connection to access database.
     * @param documentId the given document id to get the use count.
     *
     * @return the use count from the database records with given document id.
     *
     * @throws SQLException if a database error occurs.
     */
    public static int getDocumentUseCount(Connection connection, String documentId) throws SQLException {
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {
            statement = connection.prepareStatement(RETRIEVE_USECOUNT_SQL);
            statement.setString(1, documentId);
            resultSet = statement.executeQuery();

            if (!resultSet.next()) {
                return -1;
            }

            return resultSet.getInt(USE_COUNT_COLUMN);
        } finally {
            closeResources(resultSet, statement);
        }
    }

    /**
     * <p>
     * Closes JDBC resources.
     * </p>
     *
     * @param resultSet the result set to be closed.
     * @param statement the SQL statement to be closed.
     */
    private static void closeResources(ResultSet resultSet, Statement statement) {
        if (resultSet != null) {
            try {
                resultSet.close();
            } catch (SQLException se) {
                // ignore
            }
        }

        if (statement != null) {
            try {
                statement.close();
            } catch (SQLException se) {
                // ignore
            }
        }
    }
}
