/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.DocumentIndexerPersistenceHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

import com.topcoder.util.compression.Base64Codec;
import com.topcoder.util.compression.Base64Decoder;
import com.topcoder.util.compression.Base64Encoder;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;

import java.text.CollationKey;
import java.text.Collator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.zip.DataFormatException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;


/**
 * <p>
 * Implements the <code>AbstractDBIndexPersistence</code> which uses xml files as storage medium. This implementation
 * uses a dual strategy for parsing and utilized SAX parsing for reading data in and DOM parsing for data updates
 * (such as collection index updates and document use count updates). This implementation allows for creation of
 * arbitrarily large document index files by sequencing (splitting) large files into chunks of file based data. Since
 * it is required that binary data is stored in the xml files for document index a serialized Base64 encoded data will
 * be both encoded and decoded by this implementation.
 * </p>
 *
 * <p>
 * Thread-safe: this class is not thread safe.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public class XmlIndexPersistence implements IndexPersistence {
    /**
     * <p>
     * Represents the file name used to represent a document usage as defined by the persistence contract. There is
     * only one such file per xml persistence name space. This file should be insured to exist, otherwise this
     * component will make no sense.
     * </p>
     *
     * <p>
     * Here is a sample of the file:
     * <pre>
     * &lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;<br>
     *  &lt;documents&gt;<br>
     *  &lt;document-usage document-id=&quot;8849a6466364db05966ed7f8bb32dd5e&quot; usage-count=&quot;1&quot;&gt;<br>
     *  &lt;collection-inclusion&gt;<br>
     *  &lt;collection collection-id=&quot;12345&quot;/&gt;<br>
     *  &lt;/collection-inclusion&gt;<br>
     *  &lt;/document-usage&gt;<br>
     *  &lt;/documents&gt;
     * </pre>
     * </p>
     */
    private static final String DOC_USAGE_FILE_NAME = "document_usage.xml";

    /**
     * <p>
     * Represents the file name prefix for all files that represent a collection index xml file storage (i.e. a
     * document collecltion file). Each xml file name will have the following format:
     * doc_collection_collectionId_.xml.
     * </p>
     *
     * <p>
     * Here is a sample of the file:
     * <pre>
     * &lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;<br>
     *  &lt;document-collection collection-id=&quot;12345&quot;&gt;<br>
     *  &lt;documents&gt;<br>
     *  &lt;document-id&gt;8849a6466364db05966ed7f8bb32dd5e&lt;/document-id&gt;<br>
     *  &lt;/documents&gt;<br>
     *  &lt;/document-collection&gt;
     * </pre>
     * </p>
     */
    private static final String DOC_COLLECTION_FILE_NAME_PREFIX = "doc_collection_";

    /**
     * <p>
     * Represents the file name prefix for all files that represent a document index xml file storage (i.e. a document
     * index file). Each document index file name will have the following format: doc_index_documentId_sequenceid.xml.
     * The sequenceid is used to link files with data that is larger than say a limit of 2G (this can be
     * configured). This means that if we have indexed data that is larger than say 2G we split the file into 2 files
     * with the first having the sequence id of 1 and the second one being obviously 2.
     * </p>
     *
     * <p>
     * Here is a sample of the file:
     * <pre>
     * <p>&lt;doc xml:lang=&quot;en_US&quot;&gt;<br>
     *  &lt;document-index document-id=&quot;8849a6466364db05966ed7f8bb32dd5e&quot;&gt;<br>
     *  &lt;word-source-data&gt;<br>
     *  &lt;source-identity&gt;rO0ABXQAFHRlc3RfZmlsZXMvUHJlc3MudHh0&lt;/source-identity&gt;<br>
     *  &lt;class-name&gt;com.topcoder.document.index.wordsource.textfile.TextFileWordSource&lt;/class-name&gt;<br>
     *  &lt;locale language=&quot;en&quot; country=&quot;US&quot; variant=&quot;&quot; /&gt;<br>
     *  &lt;delimiters&gt;<br>
     *  &lt;entry delimiter=&quot;.&quot;/&gt;<br>
     *  &lt;entry delimiter=&quot;!&quot;/&gt;<br>
     *  &lt;entry delimiter=&quot;-&quot;/&gt;<br>
     *  &lt;/delimiters&gt;<br>
     *  &lt;/word-source-data&gt;<br>
     *  &lt;word-index&gt;<br>
     *  &lt;word&gt;&lt;![CDATA[School]]&gt;&lt;/word&gt;<br>
     *  &lt;pos&gt;27&lt;/pos&gt;<br>
     *  &lt;pos&gt;190&lt;/pos&gt;<br>
     *  &lt;/word-index&gt;<br>
     *  &lt;/document-index&gt;<br>
     *  &lt;/doc&gt;<br>
     * </pre>
     * </p>
     */
    private static final String DOC_INDEX_FILE_NAME_PREFIX = "doc_index_";

    /** Represents the suffix of the xml file. */
    private static final String XML_SUFFIX = ".xml";

    /** Represents the line separator. */
    private static final String LINE_SEPARATOR = System.getProperty("line.separator");

    /** Represents the document-usage property node in the file of document_usage.xml. */
    private static final String DOCUMENT_USAGE_PROPERTY = "document-usage";

    /** Represents the collection-inclusion property node in the file of document_usage.xml. */
    private static final String COLLECTION_INCLUSION_PROPERTY = "collection-inclusion";

    /** Represents the collection property node in the file of document_usage.xml. */
    private static final String COLLECTION_PROPERTY = "collection";

    /** Represents the document-id attribute name in the file of document_usage.xml. */
    private static final String DOCUMENT_ID_ATTRIBUTE = "document-id";

    /** Represents the usage-count attribute name in the file of document_usage.xml. */
    private static final String USAGE_COUNT_ATTRIBUTE = "usage-count";

    /** Represents the document-collection property node in the file of doc_collection_collecionId.xml. */
    private static final String DOCUMENT_COLLECTION_PROPERTY = "document-collection";

    /** Represents the document-id property name in the file of doc_collection_collecionId.xml. */
    private static final String DOCUMENT_ID_PROPERTY = "document-id";

    /** Represents the documents property node in the file of doc_collection_collecionId.xml. */
    private static final String DOCUMENTS_PROPERTY = "documents";

    /** Represents the collection-id attribute name. */
    private static final String COLLECTION_ID_ATTRIBUTE = "collection-id";

    /** Represents the count of bytes for one mega byte. */
    private static final int M = 1024 * 1024;

    /** Represents the count of bytes for one kilo byte. */
    private static final int K = 1024;

    /**
     * <p>
     * Represents a file size limit for saving a chunk of indexed data. It is measured in Mbytes. This size limit is
     * used to decide when to start chaining multiple files to store the data(since some data cannot fit into a single
     * file due to size limitations for the specific OS).
     * </p>
     */
    private final int fileSizeLimit;

    /**
     * <p>
     * Represents a safety net size that is applied before the fileSizeLimit is reached. This is basically something of
     * a round off error precaution. In effect it subtracts from the filesizeThreshold value.
     * </p>
     */
    private final int fileSizeThreshold;

    /**
     * <p>
     * Represents the file path to a directory where all the xml files that make up this persistence session are
     * stored. This can be a valid fully qualified path or a relative path.
     * </p>
     */
    private final String persistenceDirPath;

    /**
     * <p>
     * Represents the file to a directory where all the xml files that make up this persistence session are stored.
     * </p>
     */
    private final File persistenceDir;

    /**
     * <p>
     * Creates a new instance of <code>XmlIndexPersistence</code> class with given namespace.
     * </p>
     *
     * <p>
     * The property values of fileSizeLimit(required), fileSizeThreshold(required), and persistenceDirPath(required)
     * will be loaded from the configuration.
     * </p>
     *
     * @param namespace the namespace to load the property values required by this class.
     *
     * @throws IllegalArgumentException if <code>namespace</code> is <code>null</code> or empty string.
     * @throws PersistenceConfigurationException if it fails to load the property values required by this class or some
     *         of the property values is invalid.
     */
    public XmlIndexPersistence(String namespace) throws PersistenceConfigurationException {
        DocumentIndexerPersistenceHelper.validateString(namespace, "namespace");

        // retrieve the property values from the config file.
        int fileSizeLimitValue = DocumentIndexerPersistenceHelper.getIntProperty(namespace, "FileSizeLimit");
        int fileSizeThresholdValue = DocumentIndexerPersistenceHelper.getIntProperty(namespace, "FileSizeThreshold");
        this.persistenceDirPath = DocumentIndexerPersistenceHelper.getStringProperty(namespace, "XmlPersistencePath",
                true);

        // validate the property value
        if (fileSizeLimitValue <= 0) {
            throw new PersistenceConfigurationException("The value of fileSizeLimit should be positive.");
        }

        if (fileSizeThresholdValue < 0) {
            throw new PersistenceConfigurationException("The value of fileSizeThreshold should be non-negative.");
        }

        if (fileSizeLimitValue > (Integer.MAX_VALUE / M)) {
            throw new PersistenceConfigurationException("The value of fileSizeLimit is too large.");
        }

        if (fileSizeThresholdValue > (Integer.MAX_VALUE / K)) {
            throw new PersistenceConfigurationException("The value of fileSizeThreshold is too large.");
        }

        if (persistenceDirPath.trim().length() == 0) {
            throw new PersistenceConfigurationException("The value of persistenceDirPath should not be empty.");
        }

        this.fileSizeLimit = fileSizeLimitValue * M;
        this.fileSizeThreshold = fileSizeThresholdValue * K;
        this.persistenceDir = new File(persistenceDirPath);

        if (!persistenceDir.exists()) {
            throw new PersistenceConfigurationException("The persistenceDir should exist.");
        }

        if (!persistenceDir.isDirectory()) {
            throw new PersistenceConfigurationException("The persistenceDir should be a directory.");
        }

        if (!persistenceDir.canWrite()) {
            throw new PersistenceConfigurationException("The persistenceDir is not writable.");
        }
    }

    /**
     * <p>
     * Creates a new instance of <code>XmlIndexPersistence</code> class with given fileSizeLimit, fileSizeThreshold and
     * persistenceDirPath information.
     * </p>
     *
     * @param fileSizeLimit the file size(in Mbytes) limit for saving a chunk of indexed data.
     * @param fileSizeThreshold the safety net size(in Kbytes) that is applied before the fileSizeLimit is reached.
     * @param persistenceDirPath the file path to a directory where all the xml files that make up this persistence
     *        session are stored.
     *
     * @throws IllegalArgumentException if <code>persistenceDirPath</code> is <code>null</code> or empty string; or
     *         <code>fileSizeLimit</code> is not a positive integer or too large; or <code>fileSizeThreshold</code> is
     *         not a non-negative integer or too large; or <code>persistenceDirPath</code> is not a valid existing
     *         directory which is writable.
     */
    public XmlIndexPersistence(int fileSizeLimit, int fileSizeThreshold, String persistenceDirPath) {
        if (fileSizeLimit <= 0) {
            throw new IllegalArgumentException("The value of fileSizeLimit should be positive.");
        }

        if (fileSizeThreshold < 0) {
            throw new IllegalArgumentException("The value of fileSizeThreshold should be non-negative.");
        }

        if (fileSizeLimit > (Integer.MAX_VALUE / M)) {
            throw new IllegalArgumentException("The value of fileSizeLimit is too large.");
        }

        if (fileSizeThreshold > (Integer.MAX_VALUE / K)) {
            throw new IllegalArgumentException("The value of fileSizeThreshold is too large.");
        }

        if (persistenceDirPath == null) {
            throw new IllegalArgumentException("The value of persistenceDirPath should not be null.");
        }

        if (persistenceDirPath.trim().length() == 0) {
            throw new IllegalArgumentException("The value of persistenceDirPath should not be empty.");
        }

        this.fileSizeLimit = fileSizeLimit * M;
        this.fileSizeThreshold = fileSizeThreshold * K;
        this.persistenceDirPath = persistenceDirPath;
        this.persistenceDir = new File(persistenceDirPath);

        if (!persistenceDir.exists()) {
            throw new IllegalArgumentException("The persistenceDir should exist.");
        }

        if (!persistenceDir.isDirectory()) {
            throw new IllegalArgumentException("The persistenceDir should be a directory.");
        }

        if (!persistenceDir.canWrite()) {
            throw new IllegalArgumentException("The persistenceDir is not writable.");
        }
    }

    /**
     * <p>
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails.
     * </p>
     *
     * @param documentIndex the document index to be persisted.
     *
     * @throws IllegalArgumentException if <code>documentIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *         persistence.
     */
    public void addDocumentIndex(DocumentIndex documentIndex) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(documentIndex, "documentIndex");

        // get the document id
        String documentId = DocumentIndexerPersistenceHelper.getDocumentId(documentIndex.getWordSourceId());

        // check whether the document index exists in the persistence.
        if (this.getDocumentIndexFiles(documentId).length != 0) {
            throw new IndexPersistenceException("The document index exists in the persistence.");
        }

        String header = generateHeader(documentIndex, documentId);
        String footer = generateFooter();
        Map words = documentIndex.getWords();
        this.generateDocumentIndexFile(words.entrySet().iterator(), DOC_INDEX_FILE_NAME_PREFIX + documentId + "_", 1,
            header, header.getBytes().length, "", footer);

        // add the record of specific documentId into the document_usage.xml
        File originalFile = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);
        Document document = null;

        // create a DOM tree and add the specific node
        try {
            // create a new builder factory.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

            document = factory.newDocumentBuilder().parse(originalFile);

            Element root = document.getDocumentElement();

            // create the document usage node
            Element documentUsageNode = document.createElement(DOCUMENT_USAGE_PROPERTY);
            documentUsageNode.setAttribute(DOCUMENT_ID_ATTRIBUTE, documentId);
            documentUsageNode.setAttribute(USAGE_COUNT_ATTRIBUTE, "0");
            documentUsageNode.appendChild(document.createElement(COLLECTION_INCLUSION_PROPERTY));

            // append the document usage node to the root node
            root.appendChild(documentUsageNode);
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // update the content of the file
        this.commit(document, originalFile);
    }

    /**
     * Gets all the xml files which store the specific document index with the given document index id. The names of
     * all the retrieved files should be in the same format like: doc_index_docId_sequenceid.xml. All the retrieved
     * files will be sorted via its sequence number.
     *
     * @param documentId the document index id of the specific document index.
     *
     * @return the xml files which store the specific document index with the given document index id.
     *
     * @throws IndexPersistenceException if it fails to retrieve the needed xml files.
     */
    private File[] getDocumentIndexFiles(String documentId) throws IndexPersistenceException {
        File[] files = persistenceDir.listFiles();
        String prefix = DOC_INDEX_FILE_NAME_PREFIX + documentId + "_";

        List documentIndexFiles = new ArrayList();

        for (int i = 0; i < files.length; i++) {
            // get the file's name(not full name)
            String fileName = files[i].getName();

            if (fileName.startsWith(prefix) && fileName.endsWith(XML_SUFFIX)) {
                String sequenceId = fileName.substring(prefix.length());
                sequenceId = sequenceId.substring(0, sequenceId.length() - XML_SUFFIX.length());
                documentIndexFiles.add(new MyFile(files[i], this.parseInt(sequenceId)));
            }
        }

        // sort all the files via their sequence numbers.
        Collections.sort(documentIndexFiles);

        File[] ret = new File[documentIndexFiles.size()];
        int index = 0;

        // if none of the files is missing, the sequence numbers should be in this sequence: 1, 2, 3....TOTAL_SIZE,
        // otherwise it means some of the files is missing.
        for (Iterator iter = documentIndexFiles.iterator(); iter.hasNext();) {
            MyFile myFile = ((MyFile) iter.next());
            ret[index++] = myFile.file;

            if (myFile.sequenceId != index) {
                throw new IndexPersistenceException("The file with the name of " + prefix + index + " is missing.");
            }
        }

        return ret;
    }

    /**
     * <p>
     * Generates the header and all the non-word related tags.
     * </p>
     *
     * @param documentIndex the document index to store into the persistence.
     * @param documentId the id of the document index.
     *
     * @return the header and all the non-word related tags.
     *
     * @throws IndexPersistenceException any exception during the generation process.
     */
    private String generateHeader(DocumentIndex documentIndex, String documentId) throws IndexPersistenceException {
        StringBuffer buffer = new StringBuffer();
        Locale locale = documentIndex.getLocale();
        WordSourceId wordSourceId = documentIndex.getWordSourceId();
        buffer.append("<doc xml:lang=\"").append(locale.toString()).append("\">").append(LINE_SEPARATOR);
        buffer.append("\t<document-index document-id=\"").append(documentId).append("\">").append(LINE_SEPARATOR);
        buffer.append("\t\t<word-source-data>").append(LINE_SEPARATOR);
        buffer.append("\t\t\t<source-identity>").append(this.sourceIdentity2String(wordSourceId.getSourceIdentity()))
              .append("</source-identity>").append(LINE_SEPARATOR);
        buffer.append("\t\t\t<class-name>").append(wordSourceId.getSourceClassName()).append("</class-name>")
              .append(LINE_SEPARATOR);
        buffer.append("\t\t\t<locale language=\"").append(locale.getLanguage()).append("\" country=\"")
              .append(locale.getCountry()).append("\" variant=\"").append(locale.getVariant()).append("\" />")
              .append(LINE_SEPARATOR);
        buffer.append("\t\t\t<delimiters>").append(LINE_SEPARATOR);

        for (int i = 0; i < wordSourceId.getDelimiters().length; i++) {
            String delimiter = wordSourceId.getDelimiters()[i];

            // convert the special characters
            delimiter = DocumentIndexerPersistenceHelper.convertString(delimiter);

            buffer.append("\t\t\t\t<entry delimiter=\"").append(delimiter).append("\"/>").append(LINE_SEPARATOR);
        }

        buffer.append("\t\t\t</delimiters>").append(LINE_SEPARATOR);
        buffer.append("\t\t</word-source-data>").append(LINE_SEPARATOR);

        return buffer.toString();
    }

    /**
     * Converts the object of source identity into a representation of string value(base64 encoded).
     *
     * @param sourceIdentity the object of source identity to be converted.
     *
     * @return the object of source identity into a representation of string value(base64 encoded).
     *
     * @throws IndexPersistenceException any exception during the convertion.
     */
    private String sourceIdentity2String(Object sourceIdentity) throws IndexPersistenceException {
        // create the encoder
        Base64Encoder encoder = new Base64Encoder(0, null, Base64Codec.STANDARD_ALPHABET);

        // output buffer large enough to hold all the output bytes
        final int bufferSize = 1024;
        byte[] largeBuffer = new byte[bufferSize];

        // retrieve the serialized bytes
        byte[] inputBytes = DocumentIndexerPersistenceHelper.getSerializedBytes(sourceIdentity);

        // do the encoding
        encoder.setInput(inputBytes);
        encoder.finish();

        int bytesEncoded;
        int total = 0; // total bytes encoded

        while ((bytesEncoded = encoder.deflate(largeBuffer)) > 0) {
            total += bytesEncoded;
        }

        // return the encoded result
        return new String(largeBuffer, 0, total);
    }

    /**
     * <p>
     * Generates the necessary closing tags.
     * </p>
     *
     * @return the necessary closing tags.
     */
    private String generateFooter() {
        StringBuffer buffer = new StringBuffer();
        buffer.append("\t</document-index>").append(LINE_SEPARATOR);
        buffer.append("</doc>").append(LINE_SEPARATOR);

        return buffer.toString();
    }

    /**
     * <p>
     * Generates the necessary tags for the given document index. The iterator will be used to retrieve all the words
     * and their positions.
     * </p>
     *
     * <p>
     * Here is a sample for the xml persistence of document index:
     * <pre>
     *  <p>&lt;doc xml:lang=&quot;en_US&quot;&gt;<br>
     *  &nbsp;&nbsp;&nbsp; &lt;document-index document-id=&quot;1&quot;&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;word-source-data&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;source-identity&gt;null&lt;/source-identity&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;class-name&gt;com.topcoder.document.index.persistence.MockWordSource&lt;/class-name&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;locale
     *  language=&quot;en&quot; country=&quot;US&quot; variant=&quot;&quot; /&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;delimiters&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;entry delimiter=&quot; &quot;/&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;entry delimiter=&quot;,&quot;/&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;entry delimiter=&quot;?&quot;/&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;entry delimiter=&quot;!&quot;/&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;/delimiters&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;/word-source-data&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;word-index&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;word&gt;is&lt;/word&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;pos&gt;5&lt;/pos&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;pos&gt;10&lt;/pos&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;/word-index&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;word-index&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
     *  &lt;word&gt;count&lt;/word&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;pos&gt;2&lt;/pos&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;/word-index&gt;<br>
     *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &lt;word-index&gt;<br>
     *  &nbsp;&nbsp;&nbsp; &lt;/document-index&gt;<br>
     *  &lt;/doc&gt;</p>
     * </pre>
     * </p>
     *
     * @param iterator the iterator which will be used to retrieve all the words and their positions.
     * @param prefix the prefix of the xml file for document index persistence.
     * @param sequenceId the sequence id of the xml file for document index persistence.
     * @param header the header and all the non-word related tags.
     * @param headerLength the size of the bytes of the header and all the non-word related tags.
     * @param content the content of the word related tags. This value will be used since the data may be bigger than
     *        the prescribed size limit.
     * @param footer the necessary closing tags.
     *
     * @throws IndexPersistenceException if it fails to load the document index into xml persistence.
     */
    private void generateDocumentIndexFile(Iterator iterator, String prefix, int sequenceId, String header,
        int headerLength, String content, String footer) throws IndexPersistenceException {
        int currentLength = headerLength + content.getBytes().length;

        if (currentLength >= (this.fileSizeLimit - this.fileSizeThreshold)) {
            throw new IndexPersistenceException(
                "The fileSizeLimit is too small to hold a single word and its positions.");
        }

        // write the content to the file.
        BufferedWriter writer = null;
        String newContent = "";

        try {
            writer = new BufferedWriter(new FileWriter(new File(persistenceDir, prefix + sequenceId + XML_SUFFIX),
                    false));
            writer.write(header);
            writer.write(content);

            boolean hasMoreWord = false;

            while (iterator.hasNext()) {
                Map.Entry entry = (Map.Entry) iterator.next();

                // temp buffer to store the content of the word and its positions
                String tempBuffer = generateSingleWordIndex(((CollationKey) entry.getKey()).getSourceString(),
                        (List) entry.getValue());

                // calculate the current length of the output stream
                currentLength += tempBuffer.getBytes().length;

                if (currentLength >= (this.fileSizeLimit - this.fileSizeThreshold)) {
                    // data is bigger than the prescribed size limit, and we need to load it into the second file
                    newContent = tempBuffer;
                    hasMoreWord = true;

                    break;
                }

                writer.write(tempBuffer);
            }

            writer.write(footer);
            writer.flush();

            if (!hasMoreWord) {
                // no word left
                return;
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error happens in the writing process.", e);
        } finally {
            if (writer != null) {
                try {
                    writer.close();
                } catch (IOException e) {
                    // ignore here
                }
            }
        }

        // generate the next document index file
        generateDocumentIndexFile(iterator, prefix, sequenceId + 1, header, headerLength, newContent, footer);
    }

    /**
     * <p>
     * Generates the necessary tags for a single word and its positions which will be loaded into the xml persistence.
     * </p>
     *
     * @param word the word to load into the xml persistence.
     * @param positions the positions of the word to load into the xml persistence.
     *
     * @return the necessary tags for a single word and its positions which will be loaded into the xml persistence.
     */
    private String generateSingleWordIndex(String word, List positions) {
        StringBuffer buffer = new StringBuffer();
        buffer.append("\t\t<word-index>").append(LINE_SEPARATOR);

        buffer.append("\t\t\t<word>").append(word).append("</word>").append(LINE_SEPARATOR);

        for (Iterator iter = positions.iterator(); iter.hasNext();) {
            buffer.append("\t\t\t<pos>").append(((Integer) iter.next()).intValue()).append("</pos>")
            .append(LINE_SEPARATOR);
        }

        buffer.append("\t\t</word-index>").append(LINE_SEPARATOR);

        return buffer.toString();
    }

    /**
     * <p>
     * Gets the document index from the persistence. The ID of the document is given. When the document is not
     * persisted in the persistence, <code>null</code> should be returned.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is retrieved.
     *
     * @return the document index of the document with the ID.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the document index fails.
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);
        File[] files = getDocumentIndexFiles(documentId);

        if (files.length == 0) {
            // no document with given id exist
            return null;
        }

        // read all the word and position information into the words Map.
        Map words = new HashMap();

        // initialize locale-sensitive comparator
        Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());
        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < files.length; i++) {
            retrieveWordAndPositions(files[i], words, collator);
        }

        return new DocumentIndex(wordSourceId, words);
    }

    /**
     * Reads all the word and position information into an appropriate map from the specific xml persistence file.
     *
     * @param file the specific xml persistence file to load all the word and position information.
     * @param words the word map. The keys of the map are collation keys of all different words. The values are lists
     *        of integers representing the positions of the words in the document.
     * @param collator the collator instance to get the collation key.
     *
     * @throws IndexPersistenceException any exception during the parsing process.
     */
    private void retrieveWordAndPositions(File file, Map words, Collator collator) throws IndexPersistenceException {
        InputStream inputStream = null;

        try {
            // create a saxParser to parse the input file
            SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
            DefaultHandler handler = new LoadWordAndPositionsHandler();
            inputStream = new FileInputStream(file);

            // do the parsing
            parser.parse(inputStream, handler);

            // get the result word map from the handler
            Map resultWords = ((LoadWordAndPositionsHandler) handler).words;

            for (Iterator iter = resultWords.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                String word = (String) entry.getKey();

                List positions = (List) entry.getValue();

                // get the collection key
                CollationKey collectionKey = collator.getCollationKey(word);

                // re-get the positions list
                for (ListIterator listIterator = positions.listIterator(); listIterator.hasNext();) {
                    String position = (String) listIterator.next();
                    listIterator.set(new Integer(this.parseInt(position)));
                }

                // put the new word and its positions into the words map
                words.put(collectionKey, positions);
            }
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Parser cannot be created which satisfies the requested configuration.",
                e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("The underlying parser throws a SAXException while parsing.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("IO errors occur.", e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * Parse the string value into int value.
     *
     * @param stringValue the string value to parse into the int value.
     *
     * @return the parsed int value.
     *
     * @throws IndexPersistenceException if the int value is not well formatted.
     */
    private int parseInt(String stringValue) throws IndexPersistenceException {
        try {
            return Integer.parseInt(stringValue);
        } catch (NumberFormatException e) {
            throw new IndexPersistenceException("The int value of " + stringValue + " is not well formatted", e);
        }
    }

    /**
     * <p>
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The ID of the document is given. When the document is not persisted,
     * it fails.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is deleted.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *         or there is at least one collection containing the document index.
     */
    public void removeDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);
        File[] files = getDocumentIndexFiles(documentId);

        if (files.length == 0) {
            throw new IndexPersistenceException("The document index is not persisted.");
        }

        if (retrieveUseCount(documentId) != 0) {
            throw new IndexPersistenceException("There is at least one collection containing the document index.");
        }

        // delete all the indexed documents xml files
        for (int i = 0; i < files.length; i++) {
            files[i].delete();
        }

        // delete the record of specific documentId in the document_usage.xml
        File originalFile = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);
        Document document = null;

        // create a DOM tree and remove the specific node
        try {
            // create a new builder factory.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            document = factory.newDocumentBuilder().parse(originalFile);

            Element root = document.getDocumentElement();
            Element[] documentUsageNodes = this.retrieveNodesByTagName(root.getChildNodes(), DOCUMENT_USAGE_PROPERTY);

            for (int j = 0; j < documentUsageNodes.length; j++) {
                if (documentId.equals(documentUsageNodes[j].getAttribute(DOCUMENT_ID_ATTRIBUTE))) {
                    root.removeChild(documentUsageNodes[j]);
                }
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // update the content of the file
        this.commit(document, originalFile);
    }

    /**
     * Reads the use count attribute information from the document_usage.xml for specific document index.
     *
     * @param documentId the id of the document index.
     *
     * @return the use count attribute information from the document_usage.xml for specific document index.
     *
     * @throws IndexPersistenceException any exception during the parsing process.
     */
    private int retrieveUseCount(String documentId) throws IndexPersistenceException {
        InputStream inputStream = null;

        try {
            File file = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);

            // create a saxParser to parse the input file
            SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
            DefaultHandler handler = new LoadUsageCountHandler(documentId);
            inputStream = new FileInputStream(file);

            // do the parsing
            parser.parse(inputStream, handler);

            // get the usage count from the handler
            String usageCount = ((LoadUsageCountHandler) handler).usageCount;

            return this.parseInt(usageCount);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Parser cannot be created which satisfies the requested configuration.",
                e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("The underlying parser throws a SAXException while parsing.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("IO errors occur.", e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * <p>
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails.
     * </p>
     *
     * @param collectionIndex the collection index to be persisted.
     *
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in
     *         the persistence.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(collectionIndex, "collectionIndex");

        String collectionId = collectionIndex.getId();
        File originalFile = new File(this.persistenceDir, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + XML_SUFFIX);

        if (originalFile.exists()) {
            throw new IndexPersistenceException("The collection index exists in the persistence.");
        }

        Document document = null;

        // 1: create a DOM tree and add the record into doc_collection_collectionId.xml file.
        try {
            // create the document via DOM
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

            // create the root element with the specific collection id as its attribute
            Element root = document.createElement(DOCUMENT_COLLECTION_PROPERTY);
            root.setAttribute(COLLECTION_ID_ATTRIBUTE, collectionId);
            document.appendChild(root);

            // create all the other documents nodes
            Element documentsNode = document.createElement(DOCUMENTS_PROPERTY);

            for (Iterator iter = collectionIndex.getAllDocumentIds().iterator(); iter.hasNext();) {
                String documentId = DocumentIndexerPersistenceHelper.getDocumentId((WordSourceId) iter.next());
                Element documentIdNode = document.createElement(DOCUMENT_ID_PROPERTY);
                documentIdNode.appendChild(document.createTextNode(documentId));
                documentsNode.appendChild(documentIdNode);
            }

            root.appendChild(documentsNode);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // overwrite the original file with the new DOM
        this.commit(document, originalFile);

        // 2. update the records in the document_usage.xml file
        Set documentIds = new HashSet();

        for (Iterator iter = collectionIndex.getAllDocumentIds().iterator(); iter.hasNext();) {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId((WordSourceId) iter.next());

            if (this.getDocumentIndexFiles(documentId).length == 0) {
                throw new IndexPersistenceException("The document index for id: " + documentId
                    + " has not been persistenced.");
            }

            documentIds.add(documentId);
        }

        originalFile = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);
        document = null;

        // create a DOM tree and add the specific node
        try {
            // create a new builder factory.
            // create a new builder factory.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

            document = factory.newDocumentBuilder().parse(originalFile);

            Element root = document.getDocumentElement();

            // update the document usage node
            Element[] documentUsageNodes = this.retrieveNodesByTagName(root.getChildNodes(), DOCUMENT_USAGE_PROPERTY);

            for (int i = 0; i < documentUsageNodes.length; i++) {
                String documentId = documentUsageNodes[i].getAttribute(DOCUMENT_ID_ATTRIBUTE);

                if (documentIds.contains(documentId)) {
                    Element collectionInclusionNode = this.retrieveNodeByTagName(documentUsageNodes[i].getChildNodes(),
                            COLLECTION_INCLUSION_PROPERTY);

                    // add the collection node
                    Element collectionNode = document.createElement(COLLECTION_PROPERTY);
                    collectionNode.setAttribute(COLLECTION_ID_ATTRIBUTE, collectionId);
                    collectionInclusionNode.appendChild(collectionNode);
                }
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // update the content of the file
        this.commit(document, originalFile);
    }

    /**
     * <p>
     * Transfers any cached changes to permanent storage. When this method returns, any and all changes cached before
     * it was invoked will have been transferred to the permanent store.
     * </p>
     *
     * @param document the xml document which stores the cached changes.
     * @param originalFile the file to stroe the cached changes.
     *
     * @throws IndexPersistenceException any exception during the transfer.
     */
    private void commit(Document document, File originalFile) throws IndexPersistenceException {
        File tmpFile = null;

        try {
            // create the temp file.
            tmpFile = File.createTempFile("tmp", XML_SUFFIX);

            // create the transformer
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            transformer.transform(new DOMSource(document), new StreamResult(tmpFile));

            // delete the original file.
            originalFile.delete();

            // Rename the temp file to the original file.
            tmpFile.renameTo(originalFile);
        } catch (TransformerConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during saving the file.", e);
        } catch (TransformerException e) {
            throw new IndexPersistenceException("Error occurs during saving the file.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs during saving the file.", e);
        } catch (TransformerFactoryConfigurationError e) {
            throw new IndexPersistenceException("Error occurs during saving the file.", e);
        } finally {
            // delete the temp file.
            if ((tmpFile != null) && tmpFile.exists()) {
                tmpFile.delete();
            }
        }
    }

    /**
     * <p>
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, <code>null</code> should be returned.
     * </p>
     *
     * @param collectionId the ID of the collection index to be retrieved.
     *
     * @return the collection index with the given ID.
     *
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    public CollectionIndex getCollectionIndex(String collectionId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateString(collectionId, "collectionId");

        File originalFile = new File(this.persistenceDir, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + XML_SUFFIX);

        if (!originalFile.exists()) {
            return null;
        }

        String[] ids = retrieveDocumentIds(collectionId);
        Map words = new HashMap();
        Set allDocumentIds = new HashSet();

        for (int i = 0; i < ids.length; i++) {
            File[] files = getDocumentIndexFiles(ids[i]);

            if (files.length == 0) {
                throw new IndexPersistenceException("The document index for id: " + ids[i]
                    + " has not been persistenced.");
            }

            // read all the word and position information into the words Map.
            Map tempWords = new HashMap();

            Collator collator = null;
            WordSourceId wordSourceId = null;

            for (int j = 0; j < files.length; j++) {
                if (wordSourceId == null) {
                    // load the wordSourceId from the first sequence file
                    wordSourceId = this.retrieveWordSourceId(files[j]);

                    Locale locale = wordSourceId.getSourceLocale();

                    // initialize locale-sensitive comparator
                    collator = Collator.getInstance(locale);
                    collator.setStrength(Collator.PRIMARY);
                }

                retrieveWordAndPositions(files[j], tempWords, collator);
            }

            // Add all words into the word map
            for (Iterator iter = tempWords.keySet().iterator(); iter.hasNext();) {
                CollationKey key = (CollationKey) iter.next();
                Set documentIds = (Set) words.get(key);

                if (documentIds == null) {
                    // The word does not exist, create a new list
                    documentIds = new HashSet();
                    words.put(key, documentIds);
                }

                // Mark the word's existence in the document
                documentIds.add(wordSourceId);
            }

            // add the wordSourceId
            allDocumentIds.add(wordSourceId);
        }

        return new CollectionIndex(null, words, allDocumentIds, collectionId);
    }

    /**
     * Reads the use document ids from the doc_collection_collectionId.xml file for specific collection id.
     *
     * @param collectionId the id of the collection.
     *
     * @return the use document ids from the doc_collection_collectionId.xml file for specific collection id.
     *
     * @throws IndexPersistenceException any exception during the parsing process.
     */
    private String[] retrieveDocumentIds(String collectionId) throws IndexPersistenceException {
        InputStream inputStream = null;

        try {
            File file = new File(this.persistenceDir, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + XML_SUFFIX);

            // create a saxParser to parse the input file
            SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
            DefaultHandler handler = new LoadDocumentIdsHandler();
            inputStream = new FileInputStream(file);

            // do the parsing
            parser.parse(inputStream, handler);

            // get the document ids from the handler
            List documentIds = ((LoadDocumentIdsHandler) handler).documentIds;

            return (String[]) documentIds.toArray(new String[documentIds.size()]);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Parser cannot be created which satisfies the requested configuration.",
                e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("The underlying parser throws a SAXException while parsing.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("IO errors occur.", e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * Retrieves the wordSource object from the document index persistence file.
     *
     * @param file the file to retrieve the wordSource object.
     *
     * @return the wordSource object loaded from the document index persistence file.
     *
     * @throws IndexPersistenceException if it fails to retrieve the wordSource object.
     */
    private WordSourceId retrieveWordSourceId(File file) throws IndexPersistenceException {
        InputStream inputStream = null;

        try {
            // create a saxParser to parse the input file
            SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
            DefaultHandler handler = new LoadWordSourceIdHandler();
            inputStream = new FileInputStream(file);

            // do the parsing
            parser.parse(inputStream, handler);

            // get the wordSourceId from the handler
            Object sourceIdentity = string2SourceIdentity(((LoadWordSourceIdHandler) handler).sourceIdentity);
            String className = ((LoadWordSourceIdHandler) handler).className;
            Locale locale = ((LoadWordSourceIdHandler) handler).locale;
            List delimitersList = ((LoadWordSourceIdHandler) handler).delimiters;
            String[] delimiters = (String[]) delimitersList.toArray(new String[delimitersList.size()]);

            return new WordSourceId(sourceIdentity, className, delimiters, locale);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Parser cannot be created which satisfies the requested configuration.",
                e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("The underlying parser throws a SAXException while parsing.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("IO errors occur.", e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * Converts the string representation of source identity into the real object(base64 decoded).
     *
     * @param sourceIdentityString the string representation of source identity.
     *
     * @return the real object of source identity(base64 decoded).
     *
     * @throws IndexPersistenceException any exception during the convertion.
     */
    private Object string2SourceIdentity(String sourceIdentityString) throws IndexPersistenceException {
        // create the decoder
        Base64Decoder decoder = new Base64Decoder(Base64Codec.IGNORE_NONE, null, Base64Codec.STANDARD_ALPHABET);

        // output buffer large enough to hold all the output bytes
        final int bufferSize = 1024;
        byte[] largeBuffer = new byte[bufferSize];

        // do the decoding
        decoder.setInput(sourceIdentityString.getBytes());

        int bytesDecoded;
        int total = 0;

        try {
            while ((bytesDecoded = decoder.inflate(largeBuffer)) > 0) {
                total += bytesDecoded;
            }

            // return the decoded result
            return DocumentIndexerPersistenceHelper.getSerializedObject(largeBuffer, 0, total);
        } catch (DataFormatException e) {
            throw new IndexPersistenceException("Error happens in the decoding.", e);
        }
    }

    /**
     * <p>
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails.
     * </p>
     *
     * @param collectionId the ID of the collection index to be deleted.
     *
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    public void removeCollectionIndex(String collectionId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateString(collectionId, "collectionId");

        File originalFile = new File(this.persistenceDir, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + ".xml");

        if (!originalFile.exists()) {
            throw new IndexPersistenceException("The collection index ID cannot be found in the persistence.");
        }

        // 1: create a DOM tree
        // and get all document ids which are related to this collection from doc_collection_collectionId.xml file.
        String[] ids = this.retrieveDocumentIds(collectionId);

        // delete the doc_collection_collectionId.xml file
        originalFile.delete();

        // 2: update the document_usage.xml file
        originalFile = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);

        Document document = null;

        try {
            // create a new builder factory.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            document = factory.newDocumentBuilder().parse(originalFile);

            Element root = document.getDocumentElement();

            // update the document usage node
            Element[] documentUsageNodes = this.retrieveNodesByTagName(root.getChildNodes(), DOCUMENT_USAGE_PROPERTY);
            Set documentIds = new HashSet(Arrays.asList(ids));

            for (int i = 0; i < documentUsageNodes.length; i++) {
                String documentId = documentUsageNodes[i].getAttribute(DOCUMENT_ID_ATTRIBUTE);

                if (documentIds.contains(documentId)) {
                    Element collectionInclusionNode = this.retrieveNodeByTagName(documentUsageNodes[i].getChildNodes(),
                            COLLECTION_INCLUSION_PROPERTY);
                    NodeList collectionNodes = collectionInclusionNode.getElementsByTagName(COLLECTION_PROPERTY);

                    // remove the node whose collection id is specific
                    for (int j = 0; j < collectionNodes.getLength(); j++) {
                        if (collectionId.equals(((Element) collectionNodes.item(j)).getAttribute(
                                        COLLECTION_ID_ATTRIBUTE))) {
                            collectionInclusionNode.removeChild(collectionNodes.item(j));
                        }
                    }
                }
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // update the content of the file
        this.commit(document, originalFile);
    }

    /**
     * <p>
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the
     * persistence, it fails.
     * </p>
     *
     * @param collectionIndex the collection index to be updated.
     *
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(collectionIndex, "collectionIndex");

        String collectionId = collectionIndex.getId();
        File originalFile = new File(this.persistenceDir, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + ".xml");

        if (!originalFile.exists()) {
            throw new IndexPersistenceException("The collection index ID cannot be found in the persistence.");
        }

        // if there is any document index entries there that are not in the input CollectionIndex object,
        // update the document usage node(remove the document index of the specific collection id)
        Set documentIdsToDelete = new HashSet(Arrays.asList(this.retrieveDocumentIds(collectionId)));

        // if there any document index entries in the CollectionIndex object that are not in the current
        // persistence, update the document usage node(add the document index of the specific collection id)
        List documentIdsToAdd = new ArrayList();

        for (Iterator iter = collectionIndex.getAllDocumentIds().iterator(); iter.hasNext();) {
            String documentId = DocumentIndexerPersistenceHelper.getDocumentId((WordSourceId) iter.next());

            if (!documentIdsToDelete.contains(documentId)) {
                // if there any document index entries in the CollectionIndex object that are not in the current
                // persistence, update the document usage node(add the document index of the specific collection id)
                documentIdsToAdd.add(documentId);
            } else {
                documentIdsToDelete.remove(documentId);
            }
        }

        Document document = null;

        // 1: create a DOM tree and add/delete the records from doc_collection_collectionId.xml file.
        try {
            // create the document via DOM
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(originalFile);

            // get the root element
            Element root = document.getDocumentElement();

            // get all the document-id nodes
            Element documentsNode = this.retrieveNodeByTagName(root.getChildNodes(), DOCUMENTS_PROPERTY);
            NodeList documentIdNodes = documentsNode.getElementsByTagName(DOCUMENT_ID_PROPERTY);

            // remove the document ids to delete
            for (int i = 0; i < documentIdNodes.getLength(); i++) {
                Element documentIdNode = (Element) documentIdNodes.item(i);
                String documentId = extractText(documentIdNode);

                if (documentIdsToDelete.contains(documentId)) {
                    documentsNode.removeChild(documentIdNode);
                }
            }

            // add the document ids to add
            for (Iterator iter = documentIdsToAdd.iterator(); iter.hasNext();) {
                String documentId = (String) iter.next();
                Element documentIdNode = document.createElement(DOCUMENT_ID_PROPERTY);
                documentIdNode.appendChild(document.createTextNode(documentId));
                documentsNode.appendChild(documentIdNode);
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // overwrite the original file with the new DOM
        this.commit(document, originalFile);

        // 2. create a DOM tree to update the records in the document_usage.xml file
        originalFile = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);
        document = null;

        // create a DOM tree and add/delete the specific node
        try {
            // create a new builder factory.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

            document = factory.newDocumentBuilder().parse(originalFile);

            Element root = document.getDocumentElement();

            // update the document usage node
            Element[] documentUsageNodes = this.retrieveNodesByTagName(root.getChildNodes(), DOCUMENT_USAGE_PROPERTY);

            for (int i = 0; i < documentUsageNodes.length; i++) {
                String documentId = documentUsageNodes[i].getAttribute(DOCUMENT_ID_ATTRIBUTE);

                if (documentIdsToAdd.contains(documentId)) {
                    Element collectionInclusionNode = this.retrieveNodeByTagName(documentUsageNodes[i].getChildNodes(),
                            COLLECTION_INCLUSION_PROPERTY);

                    // add the collection node
                    Element collectionNode = document.createElement(COLLECTION_PROPERTY);
                    collectionNode.setAttribute(COLLECTION_ID_ATTRIBUTE, collectionId);
                    collectionInclusionNode.appendChild(collectionNode);
                } else if (documentIdsToDelete.contains(documentId)) {
                    Element collectionInclusionNode = this.retrieveNodeByTagName(documentUsageNodes[i].getChildNodes(),
                            COLLECTION_INCLUSION_PROPERTY);
                    NodeList collectionNodes = collectionInclusionNode.getElementsByTagName(COLLECTION_PROPERTY);

                    // remove the node whose collection id is specific
                    for (int j = 0; j < collectionNodes.getLength(); j++) {
                        if (collectionId.equals(((Element) collectionNodes.item(j)).getAttribute(
                                        COLLECTION_ID_ATTRIBUTE))) {
                            collectionInclusionNode.removeChild(collectionNodes.item(j));
                        }
                    }
                }
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // update the content of the file
        this.commit(document, originalFile);
    }

    /**
     * <p>
     * A utility method for extracting the text content of a DOM Node, accounting for Text nodes, CDATA sections, and
     * intervening nodes of other types.
     * </p>
     *
     * @param node the Node to extract the string value.
     *
     * @return a String containing the text content of the specified Node.
     */
    private String extractText(Node node) {
        // using StringBuffer to append all the text.
        StringBuffer buffer = new StringBuffer();

        // iterate through all the text or CDATA children.
        NodeList childNodes = node.getChildNodes();

        for (int i = 0; i < childNodes.getLength(); i++) {
            Node child = childNodes.item(i);

            // Only CDATA_SECTION_NODE and TEXT_NODE will be added to the result.
            if ((child.getNodeType() == Node.CDATA_SECTION_NODE) || (child.getNodeType() == Node.TEXT_NODE)) {
                buffer.append(child.getNodeValue());
            }
        }

        return buffer.toString();
    }

    /**
     * <p>
     * Increases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);
        File[] files = getDocumentIndexFiles(documentId);

        if (files.length == 0) {
            // no document with given id exist
            throw new IndexPersistenceException("The document index is not persisted.");
        }

        File originalFile = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);
        Document document = null;

        // create a DOM tree and update the specific node
        try {
            // create a new builder factory.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            document = factory.newDocumentBuilder().parse(originalFile);

            Element root = document.getDocumentElement();
            Element[] documentUsageNodes = this.retrieveNodesByTagName(root.getChildNodes(), DOCUMENT_USAGE_PROPERTY);

            for (int j = 0; j < documentUsageNodes.length; j++) {
                if (documentId.equals(documentUsageNodes[j].getAttribute(DOCUMENT_ID_ATTRIBUTE))) {
                    try {
                        int preCount = Integer.parseInt(documentUsageNodes[j].getAttribute(USAGE_COUNT_ATTRIBUTE));
                        ++preCount;
                        documentUsageNodes[j].setAttribute(USAGE_COUNT_ATTRIBUTE, "" + preCount);
                    } catch (NumberFormatException e) {
                        throw new IndexPersistenceException("The usage-count is not properly formatted.", e);
                    }
                }
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // update the content of the file
        this.commit(document, originalFile);
    }

    /**
     * <p>
     * Decreases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *         not persisted; or the collection reference count is already 0.
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);
        File[] files = getDocumentIndexFiles(documentId);

        if (files.length == 0) {
            // no document with given id exist
            throw new IndexPersistenceException("The document index is not persisted.");
        }

        File originalFile = new File(this.persistenceDir, DOC_USAGE_FILE_NAME);
        Document document = null;

        // create a DOM tree and update the specific node
        try {
            // create a new builder factory.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            document = factory.newDocumentBuilder().parse(originalFile);

            Element root = document.getDocumentElement();
            Element[] documentUsageNodes = this.retrieveNodesByTagName(root.getChildNodes(), DOCUMENT_USAGE_PROPERTY);

            for (int j = 0; j < documentUsageNodes.length; j++) {
                if (documentId.equals(documentUsageNodes[j].getAttribute(DOCUMENT_ID_ATTRIBUTE))) {
                    try {
                        int preCount = Integer.parseInt(documentUsageNodes[j].getAttribute(USAGE_COUNT_ATTRIBUTE));

                        if (preCount == 0) {
                            throw new IndexPersistenceException("The collection reference count is already 0.");
                        }

                        --preCount;
                        documentUsageNodes[j].setAttribute(USAGE_COUNT_ATTRIBUTE, "" + preCount);
                    } catch (NumberFormatException e) {
                        throw new IndexPersistenceException("The usage-count is not properly formatted.", e);
                    }
                }
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("Error occurs when parsing file.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        } catch (DOMException e) {
            throw new IndexPersistenceException("Error occurs during creating Document for file.", e);
        }

        // update the content of the file
        this.commit(document, originalFile);
    }

    /**
     * <p>
     * Gets a set of all IDs of documents whose indices are persisted in the persistence.
     * </p>
     *
     * @return a set of all IDs of persisted document indices.
     *
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        File[] files = getDocumentIndexFiles();
        Set ret = new HashSet();

        for (int i = 0; i < files.length; i++) {
            ret.add(this.retrieveWordSourceId(files[i]));
        }

        return ret;
    }

    /**
     * <p>
     * Gets all the document index files whose name is in the format of doc_index_documentId_1.xml.
     * </p>
     *
     * @return all the document index files whose name is in the format of doc_index_documentId_1.xml.
     *
     * @throws IndexPersistenceException if it fails to get the document index files.
     */
    private File[] getDocumentIndexFiles() throws IndexPersistenceException {
        File[] files = persistenceDir.listFiles();
        String ends = "_1.xml";
        List documentIndexFiles = new ArrayList();

        for (int i = 0; i < files.length; i++) {
            String fileName = files[i].getName();

            if (fileName.startsWith(DOC_INDEX_FILE_NAME_PREFIX) && fileName.endsWith(ends)) {
                documentIndexFiles.add(files[i]);
            }
        }

        return (File[]) documentIndexFiles.toArray(new File[documentIndexFiles.size()]);
    }

    /**
     * <p>
     * Gets the reference count of the given document index. When the document is not in the persistence, return 0.
     * </p>
     *
     * @param wordSourceId the ID of the document index whose reference count is retrieved.
     *
     * @return the reference count of the document index.
     *
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *         not persisted.
     */
    public int getDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        DocumentIndexerPersistenceHelper.validateNotNull(wordSourceId, "wordSourceId");

        String documentId = DocumentIndexerPersistenceHelper.getDocumentId(wordSourceId);

        return this.retrieveUseCount(documentId);
    }

    /**
     * <p>
     * Retrieve the xml node by the specific tag name.
     * </p>
     *
     * @param nodes the nodelist where to retrieve the required xml node.
     * @param name the specific tag name.
     *
     * @return the xml node by the specific tag name.
     */
    private Element retrieveNodeByTagName(NodeList nodes, String name) {
        // Not using Element.getElementsByTagName() here since this method will iterate all descendant Elements with a
        // given tag name, so it is not efficient.
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);

            if ((node.getNodeType() == Node.ELEMENT_NODE) && node.getNodeName().equals(name)) {
                return (Element) node;
            }
        }

        return null;
    }

    /**
     * <p>
     * Retrieve the xml nodes by the specific tag name.
     * </p>
     *
     * @param nodes the nodelist where to retrieve the required xml node.
     * @param name the specific tag name.
     *
     * @return the xml nodes by the specific tag name.
     */
    private Element[] retrieveNodesByTagName(NodeList nodes, String name) {
        // Not using Element.getElementsByTagName() here since this method will iterate all descendant Elements with a
        // given tag name, so it is not efficient.
        List nodesList = new ArrayList();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);

            if ((node.getNodeType() == Node.ELEMENT_NODE) && node.getNodeName().equals(name)) {
                nodesList.add((Element) node);
            }
        }

        return (Element[]) nodesList.toArray(new Element[nodesList.size()]);
    }

    /**
     * <p>
     * A helper class to store the file and its sequence id. This class will be used for sorting purpose.
     * </p>
     *
     * @author TCSDEVELOPER
     * @version 1.0
     */
    private class MyFile implements Comparable {
        /** Represents the file. */
        private File file = null;

        /** Represents the sequence id of the file. */
        private int sequenceId = 0;

        /**
         * Creates a new instance of <code>MyFile</code> class with given file and its sequence id.
         *
         * @param file the file.
         * @param sequenceId the sequence id of the file.
         */
        public MyFile(File file, int sequenceId) {
            this.sequenceId = sequenceId;
            this.file = file;
        }

        /**
         * Overloads the <code>compareTo</code> method for sorting purpose. The sorting will be based on the sequence
         * id.
         *
         * @param obj the another <code>MyFile</code> instance.
         *
         * @return the compare result of the two <code>MyFile</code> instances.
         */
        public int compareTo(Object obj) {
            return sequenceId - ((MyFile) obj).sequenceId;
        }
    }

    /**
     * <p>
     * The handler used to parse an xml file to load the woudSourceId.
     * </p>
     *
     * @author TCSDEVELOPER
     * @version 1.0
     *
     * @copyright Copyright &copy 2006 TopCoder, Inc., All Rights Reserved
     */
    private class LoadWordSourceIdHandler extends DefaultHandler {
        /**
         * Represents whether reading value of not.
         */
        private boolean readValue = false;

        /**
         * Represents the sourceIdentity of the wordSourceId.
         */
        private String sourceIdentity = null;

        /**
         * Represents the className of the wordSourceId.
         */
        private String className = null;

        /**
         * Represents the locale of the wordSourceId.
         */
        private Locale locale = null;

        /**
         * Represents the delimiters of the wordSourceId.
         */
        private List delimiters = new ArrayList();

        /**
         * Represents the content of generated in the parsing process.
         */
        private String content = null;

        /**
         * <p>
         * The constructor.
         * </p>
         */
        public LoadWordSourceIdHandler() {
        }

        /**
         * <p>
         * Override the method to handle the start of an element.
         * </p>
         *
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace
         *        processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace processing is not being
         *        performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are not available.
         * @param attributes The attributes attached to the element. If there are no attributes, it shall be an empty
         *        Attributes object.
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) {
            if (qName.equals("source-identity") || qName.equals("class-name")) {
                readValue = true;
            } else if (qName.equals("entry")) {
                delimiters.add(attributes.getValue("delimiter"));
            } else if (qName.equals("locale")) {
                locale = new Locale(attributes.getValue("language"), attributes.getValue("country"),
                        attributes.getValue("variant"));
            }
        }

        /**
         * <p>
         * Override the method to handle the values.
         * </p>
         *
         * @param ch The characters.
         * @param start The start position in the character array.
         * @param length The number of characters to use from the character array.
         */
        public void characters(char[] ch, int start, int length) {
            if (this.readValue) {
                content = new String(ch, start, length);
                this.readValue = false;
            }
        }

        /**
         * <p>
         * Override the method to handle the values.
         * </p>
         *
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace
         *        processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace processing is not being
         *        performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are not available.
         */
        public void endElement(String uri, String localName, String qName) {
            if (qName.equals("source-identity")) {
                this.sourceIdentity = content;
            } else if (qName.equals("class-name")) {
                this.className = content;
            }
        }
    }

    /**
     * <p>
     * The handler used to parse an xml file to load the woudSourceId.
     * </p>
     *
     * @author TCSDEVELOPER
     * @version 1.0
     */
    private class LoadWordAndPositionsHandler extends DefaultHandler {
        /**
         * Represents whether reading value of not.
         */
        private boolean readValue = false;

        /**
         * Represents the word.
         */
        private String word = null;

        /**
         * Represents the word positions.
         */
        private List positions = new ArrayList();

        /**
         * Represents the content of generated in the parsing process.
         */
        private String content = null;

        /**
         * Represents the word map. The keys of the map are collation keys of all different words. The values are lists
         * of integers representing the positions of the words in the document.
         */
        private Map words = new HashMap();

        /**
         * <p>
         * The constructor.
         * </p>
         */
        public LoadWordAndPositionsHandler() {
        }

        /**
         * <p>
         * Override the method to handle the start of an element.
         * </p>
         *
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace
         *        processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace processing is not being
         *        performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are not available.
         * @param attributes The attributes attached to the element. If there are no attributes, it shall be an empty
         *        Attributes object.
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) {
            if (qName.equals("word") || qName.equals("pos")) {
                readValue = true;
            }
        }

        /**
         * <p>
         * Override the method to handle the values.
         * </p>
         *
         * @param ch The characters.
         * @param start The start position in the character array.
         * @param length The number of characters to use from the character array.
         */
        public void characters(char[] ch, int start, int length) {
            if (this.readValue) {
                content = new String(ch, start, length);
                this.readValue = false;
            }
        }

        /**
         * <p>
         * Override the method to handle the values.
         * </p>
         *
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace
         *        processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace processing is not being
         *        performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are not available.
         */
        public void endElement(String uri, String localName, String qName) {
            if (qName.equals("word")) {
                this.word = content;
            } else if (qName.equals("pos")) {
                this.positions.add(content);
            } else if (qName.equals("word-index")) {
                words.put(word, positions);
                positions = new ArrayList();
            }
        }
    }

    /**
     * <p>
     * The handler used to parse an xml file to load the document ids.
     * </p>
     *
     * @author TCSDEVELOPER
     * @version 1.0
     */
    private class LoadDocumentIdsHandler extends DefaultHandler {
        /**
         * Represents whether reading value of not.
         */
        private boolean readValue = false;

        /**
         * Represents the content of generated in the parsing process.
         */
        private String content = null;

        /**
         * Represents the document ids.
         */
        private List documentIds = new ArrayList();

        /**
         * <p>
         * The constructor.
         * </p>
         */
        public LoadDocumentIdsHandler() {
        }

        /**
         * <p>
         * Override the method to handle the start of an element.
         * </p>
         *
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace
         *        processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace processing is not being
         *        performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are not available.
         * @param attributes The attributes attached to the element. If there are no attributes, it shall be an empty
         *        Attributes object.
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) {
            if (qName.equals(DOCUMENT_ID_PROPERTY)) {
                readValue = true;
            }
        }

        /**
         * <p>
         * Override the method to handle the values.
         * </p>
         *
         * @param ch The characters.
         * @param start The start position in the character array.
         * @param length The number of characters to use from the character array.
         */
        public void characters(char[] ch, int start, int length) {
            if (this.readValue) {
                content = new String(ch, start, length);
                this.readValue = false;
            }
        }

        /**
         * <p>
         * Override the method to handle the values.
         * </p>
         *
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace
         *        processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace processing is not being
         *        performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are not available.
         */
        public void endElement(String uri, String localName, String qName) {
            if (qName.equals(DOCUMENT_ID_PROPERTY)) {
                this.documentIds.add(content);
            }
        }
    }

    /**
     * <p>
     * The handler used to parse an xml file to load the usage count of the specific document index.
     * </p>
     *
     * @author TCSDEVELOPER
     * @version 1.0
     */
    private class LoadUsageCountHandler extends DefaultHandler {
        /**
         * Represents the usage count.
         */
        private String usageCount = "-1";

        /**
         * Represents the document id whose usage count will be retrieved.
         */
        private String documentId;

        /**
         * <p>
         * The constructor.
         * </p>
         *
         * @param documentId the document id whose usage count will be retrieved.
         */
        public LoadUsageCountHandler(String documentId) {
            this.documentId = documentId;
        }

        /**
         * <p>
         * Override the method to handle the start of an element.
         * </p>
         *
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace
         *        processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace processing is not being
         *        performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are not available.
         * @param attributes The attributes attached to the element. If there are no attributes, it shall be an empty
         *        Attributes object.
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) {
            if (qName.equals(DOCUMENT_USAGE_PROPERTY)) {
                if (documentId.equals(attributes.getValue(DOCUMENT_ID_ATTRIBUTE))) {
                    usageCount = attributes.getValue(USAGE_COUNT_ATTRIBUTE);
                }
            }
        }
    }
}
