/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.IndexManager;
import com.topcoder.document.index.persistence.impl.db.FastAccessDBIndexPersistence;
import com.topcoder.document.index.persistence.impl.db.SimpleAccessDBIndexPersistence;
import com.topcoder.document.index.persistence.impl.xml.XmlIndexPersistence;
import com.topcoder.document.index.wordsource.WordSource;
import com.topcoder.document.index.wordsource.textfile.TextFileWordSource;

import junit.framework.TestCase;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;

import java.sql.Connection;

import java.util.Set;


/**
 * <p>
 * Demostrates the usage of this component. The usage includes four parts.
 * </p>
 *
 * <p>
 * The first part will introduce the usage of managering based persistence usage.
 * </p>
 *
 * <p>
 * The second part will introduce the usage of simple db persistence usage.
 * </p>
 *
 * <p>
 * The third part will introduce the usage of fast db persistence usage.
 * </p>
 *
 * <p>
 * The last part will introduce the usage of xml persistence usage.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class DemoTest extends TestCase {
    /**
     * <p>
     * This represent the file name used to represent a document usage as defined by the persistence contract. There is
     * only one such file per xml persistence name space.
     * </p>
     */
    private static final String DOC_USAGE_FILE_NAME = "document_usage.xml";

    /** Represents the directory used to store all the xml persistence files used for testing. */
    private static final String PERSISTENCE_DIR = "test_files/XmlPersistence";

    /** Represents the delimiters used to separate words. */
    private static final String[] DELIMITERS = {" ", "\n", "\t", "\r", ",", "\"", ".", "!", "-", "?", "(", ")"};

    /** Represents the namespace to load DB connection factory configuration. */
    private static final String DB_NAMESPACE = "com.topcoder.document.index.persistence.impl.db";

    /** Represents the connection name to get DB connection factory configuration. */
    private static final String CONN_NAME = "FastAccessDBIndexPersistence";

    /** Represents the database connection to access database. */
    private Connection connection;

    /** Represents a DB connection factory to create connections. */
    private DBConnectionFactory factory;

    /** Represents the <code>DocumentIndex</code> instance with large size of words used for testing. */
    private DocumentIndex documentIndex = null;

    /** Represents the <code>CollectionIndex</code> instance with small size of words used for testing. */
    private CollectionIndex collectionIndex = null;

    /** Represents the <code>CollectionIndex</code> instance with small size of words used for testing. */
    private CollectionIndex emptyCollectionIndex = null;

    /**
     * <p>
     * Sets up the test environment. Valid configurations are loaded. The test instance is created.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void setUp() throws Exception {
        UnitTestHelper.clearConfiguration();
        UnitTestHelper.addValidConfiguration();

        factory = new DBConnectionFactoryImpl(DB_NAMESPACE);
        connection = factory.createConnection(CONN_NAME);

        UnitTestHelper.clearFastDatabase(connection);

        // clear all files
        File[] files = new File(PERSISTENCE_DIR).listFiles();

        for (int i = 0; i < files.length; i++) {
            files[i].delete();
        }

        // create the document_usage.xml for testing.
        File file = new File(PERSISTENCE_DIR, DOC_USAGE_FILE_NAME);
        BufferedWriter writer = new BufferedWriter(new FileWriter(file));
        writer.write("<documents></documents>");
        writer.flush();
        writer.close();

        // Create the document index
        WordSource doc = new TextFileWordSource("test_files/Press.txt", DELIMITERS);
        documentIndex = new DocumentIndex(doc);

        // Create the collection index
        collectionIndex = UnitTestHelper.createCollectionIndex(new DocumentIndex[] {documentIndex}, "collectionID1");

        emptyCollectionIndex = UnitTestHelper.createCollectionIndex(new DocumentIndex[0], "265727");
    }

    /**
     * Removes the configuration if it exists. All the tables will be cleared.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void tearDown() throws Exception {
        UnitTestHelper.clearConfiguration();

        UnitTestHelper.clearFastDatabase(connection);

        connection.close();

        // clear all files
        File[] files = new File(PERSISTENCE_DIR).listFiles();

        for (int i = 0; i < files.length; i++) {
            files[i].delete();
        }
    }

    /**
     * Show the usage of managering based persistence usage.
     *
     * @throws Exception any exception to junit.
     */
    public void testBasicPersistenceUsage() throws Exception {
        // Create a xml based persistence
        IndexPersistence xmlPersistence = new XmlIndexPersistence(
                "com.topcoder.document.index.persistence.impl.xml.XmlIndexPersistence");

        // pass it to the manager
        IndexManager manager = new IndexManager(xmlPersistence);

        // Create a fast db based persistence
        IndexPersistence fastDBPersistence = new FastAccessDBIndexPersistence(
                "com.topcoder.document.index.persistence.impl.db.FastAccessDBIndexPersistence");

        // pass it to the manager
        manager = new IndexManager(fastDBPersistence);

        // Create a simple db based persistence
        IndexPersistence simpleDBPersistence = new SimpleAccessDBIndexPersistence(
                "com.topcoder.document.index.persistence.impl.db.SimpleAccessDBIndexPersistence");

        // pass it to the manager
        manager = new IndexManager(simpleDBPersistence);
    }

    /**
     * Show the usage of simple db access persistence usage.
     *
     * @throws Exception any exception to junit.
     */
    public void testSimpleAccessDBPersistenceUsage() throws Exception {
        // Use the API to do some specific persistence. Here we showcase the direct persistence
        // API on standalone basis rather than through the index manager.
        // We will use the SimpleAccessDBIndexPersistence instance
        // Create a fast access db based persistence
        IndexPersistence simpleAccessDBIndexPersistence = new SimpleAccessDBIndexPersistence(
                "com.topcoder.document.index.persistence.impl.db.SimpleAccessDBIndexPersistence");

        // add a document to the persistence
        simpleAccessDBIndexPersistence.addDocumentIndex(documentIndex);

        // get it back
        documentIndex = simpleAccessDBIndexPersistence.getDocumentIndex(this.documentIndex.getWordSourceId());

        // create a collection index
        simpleAccessDBIndexPersistence.addCollectionIndex(collectionIndex);

        // get some persisted collection index
        simpleAccessDBIndexPersistence.addCollectionIndex(emptyCollectionIndex);

        CollectionIndex someDocumentCollection = simpleAccessDBIndexPersistence.getCollectionIndex("265727");

        // remove it
        simpleAccessDBIndexPersistence.removeCollectionIndex("265727");

        // change something in the collection
        collectionIndex = UnitTestHelper.createCollectionIndex(new DocumentIndex[0], collectionIndex.getId());

        // update it
        simpleAccessDBIndexPersistence.updateCollectionIndex(collectionIndex);

        // increase the usage count
        simpleAccessDBIndexPersistence.increaseDocumentUseCount(documentIndex.getWordSourceId());

        // decrease the usage count
        simpleAccessDBIndexPersistence.decreaseDocumentUseCount(documentIndex.getWordSourceId());

        // get the usage count
        int persistedCount = simpleAccessDBIndexPersistence.getDocumentUseCount(documentIndex.getWordSourceId());

        // get all persisted docs
        Set docs = simpleAccessDBIndexPersistence.getIndexedDocuments();

        // remove the document index
        simpleAccessDBIndexPersistence.removeDocumentIndex(documentIndex.getWordSourceId());
    }

    /**
     * Show the usage of fast db access persistence usage.
     *
     * @throws Exception any exception to junit.
     */
    public void testFastAccessDBPersistenceUsage() throws Exception {
        // Use the API to do some specific persistence. Here we showcase the direct persistence
        // API on standalone basis rather than through the index manager. We will use the FastAccessDBIndexPersistence instance
        // Create a fast access db based persistence
        IndexPersistence fastAccessDBIndexPersistence = new FastAccessDBIndexPersistence(
                "com.topcoder.document.index.persistence.impl.db.FastAccessDBIndexPersistence");

        // add a document to the persistence
        fastAccessDBIndexPersistence.addDocumentIndex(documentIndex);

        // get it back
        documentIndex = fastAccessDBIndexPersistence.getDocumentIndex(this.documentIndex.getWordSourceId());

        // create a collection index
        fastAccessDBIndexPersistence.addCollectionIndex(collectionIndex);

        // get some persisted collection index
        fastAccessDBIndexPersistence.addCollectionIndex(emptyCollectionIndex);

        CollectionIndex someDocumentCollection = fastAccessDBIndexPersistence.getCollectionIndex("265727");

        // remove it
        fastAccessDBIndexPersistence.removeCollectionIndex("265727");

        // change something in the collection
        collectionIndex = UnitTestHelper.createCollectionIndex(new DocumentIndex[0], collectionIndex.getId());

        // update it
        fastAccessDBIndexPersistence.updateCollectionIndex(collectionIndex);

        // increase the usage count
        fastAccessDBIndexPersistence.increaseDocumentUseCount(documentIndex.getWordSourceId());

        // decrease the usage count
        fastAccessDBIndexPersistence.decreaseDocumentUseCount(documentIndex.getWordSourceId());

        // get the usage count
        int persistedCount = fastAccessDBIndexPersistence.getDocumentUseCount(documentIndex.getWordSourceId());

        // get all persisted docs
        Set docs = fastAccessDBIndexPersistence.getIndexedDocuments();

        // remove the document index
        fastAccessDBIndexPersistence.removeDocumentIndex(documentIndex.getWordSourceId());
    }

    /**
     * Show the usage of xml persistence usage.
     *
     * @throws Exception any exception to junit.
     */
    public void testXMLPersistenceUsage() throws Exception {
        // Use the API to do some specific persistence. Here we showcase the direct persistence
        // API on standalone basis rather than through the index manager. We will use the xmlPersistence instance
        // Create a xml based persistence
        IndexPersistence xmlPersistence = new XmlIndexPersistence(
                "com.topcoder.document.index.persistence.impl.xml.XmlIndexPersistence");

        // add a document to the persistence
        xmlPersistence.addDocumentIndex(documentIndex);

        // get it back
        documentIndex = xmlPersistence.getDocumentIndex(this.documentIndex.getWordSourceId());

        // create a collection index
        xmlPersistence.addCollectionIndex(collectionIndex);

        // get some persisted collection index
        xmlPersistence.addCollectionIndex(emptyCollectionIndex);

        CollectionIndex someDocumentCollection = xmlPersistence.getCollectionIndex("265727");

        // remove it
        xmlPersistence.removeCollectionIndex("265727");

        // change something in the collection
        collectionIndex = UnitTestHelper.createCollectionIndex(new DocumentIndex[0], collectionIndex.getId());

        // update it
        xmlPersistence.updateCollectionIndex(collectionIndex);

        // increase the usage count
        xmlPersistence.increaseDocumentUseCount(documentIndex.getWordSourceId());

        // decrease the usage count
        xmlPersistence.decreaseDocumentUseCount(documentIndex.getWordSourceId());

        // get the usage count
        int persistedCount = xmlPersistence.getDocumentUseCount(documentIndex.getWordSourceId());

        // get all persisted docs
        Set docs = xmlPersistence.getIndexedDocuments();

        // remove the document index
        xmlPersistence.removeDocumentIndex(documentIndex.getWordSourceId());
    }
}
