/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import com.topcoder.document.index.wordsource.WordSource;
import com.topcoder.document.index.wordsource.WordSourceException;
import com.topcoder.document.index.wordsource.WordSourceId;
import com.topcoder.document.index.wordsource.WordSourceWordInfo;

import java.util.Locale;


/**
 * <p>
 * Implements <code>WordSource</code> interface to provide the functionality to extract words from the identity. The
 * identity is an array of strings representing the words.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class MockWordSource implements WordSource {
    /** Represents the locale of the document. */
    private final Locale locale;

    /** Represents the delimiters used to separate the words. The delimiters are matched in the order of the array. */
    private final String[] delimiters;

    /** Represents the words of this document. */
    private final String[] words;

    /** Represents the current position of the words. When it is -1, the document is closed. */
    private int currentPosition = 0;

    /** Represents the ID of this document. */
    private final WordSourceId sourceId;

    /** Represents the sourceIdentity of this document. */
    private final Object sourceIdentity;

    /**
     * <p>
     * Creates a new instance of <code>MockWordSource</code> class. The identity, which should be an array of strings
     * representing the words, the delimiters of the words, and the locale of the document are given.
     * </p>
     *
     * @param identity the identity of the document content, which should be a file name.
     * @param delimiters the array of delimiters used to separate words in the document.
     * @param locale the locale of the document.
     *
     * @throws IllegalArgumentException if <code>identity</code> is not an array of strings representing the words.
     */
    public MockWordSource(Object identity, String[] delimiters, Locale locale) {
        if (!(identity instanceof String[])) {
            throw new IllegalArgumentException("The identity is not a string array.");
        }

        sourceIdentity = "words";
        words = (String[]) identity;
        sourceId = new WordSourceId(sourceIdentity, getClass().getName(), delimiters, locale);
        this.delimiters = (String[]) delimiters.clone();
        this.locale = locale;
    }

    /**
     * <p>
     * Checks if the document is already closed.
     * </p>
     *
     * @throws WordSourceException if the document is already closed.
     */
    private void checkClose() throws WordSourceException {
        if (currentPosition < 0) {
            throw new WordSourceException("The document is closed.");
        }
    }

    /**
     * <p>
     * Get a flag indicating whether there are more words in the document.
     * </p>
     *
     * @return <code>true</code> if there are more words in the document; <code>false</code> otherwise.
     *
     * @throws WordSourceException if the document is closed.
     */
    public boolean hasWords() throws WordSourceException {
        checkClose();

        return currentPosition < words.length;
    }

    /**
     * <p>
     * Gets the next word in the document using the delimiters. The starting position of the word and the word itself
     * are returned. When there is no more word, it fails.
     * </p>
     *
     * @return the next word and the starting position of the word.
     *
     * @throws WordSourceException if the document is closed; or there is no more word.
     */
    public WordSourceWordInfo nextWord() throws WordSourceException {
        if (!hasWords()) {
            throw new WordSourceException("There is no more word.");
        }

        return new WordSourceWordInfo(words[currentPosition], currentPosition++);
    }

    /**
     * <p>
     * Gets the identity object representing the content of the document.
     * </p>
     *
     * @return the identity object representing the content of the document.
     */
    public Object getSourceIdentity() {
        return sourceIdentity;
    }

    /**
     * <p>
     * Gets the locale of the document used in comparison.
     * </p>
     *
     * @return the locale of the document.
     */
    public Locale getLocale() {
        return locale;
    }

    /**
     * <p>
     * Gets the ID of the document. The ID contains identity of the document content, the locale and the word
     * delimiters.
     * </p>
     *
     * @return the ID of the document.
     */
    public WordSourceId getId() {
        return sourceId;
    }

    /**
     * <p>
     * Gets the array of word delimiters used to separate words in the document. The delimiters cannot be strings with
     * length of 0.
     * </p>
     *
     * @return the array of word delimiters.
     */
    public String[] getDelimiters() {
        return delimiters;
    }

    /**
     * <p>
     * Closes the document. No further reading operation can be done.
     * </p>
     */
    public void close() {
        currentPosition = -1;
    }
}
