/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.CollationKey;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.util.config.ConfigManager;
import com.topcoder.util.config.ConfigManagerException;


/**
 * <p>
 * Defines helper methods used in tests.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public final class UnitTestHelper {
    /**
     * <p>
     * Creates a new instance of <code>UnitTestHelper</code> class. The private constructor prevents the creation of a
     * new instance.
     * </p>
     */
    private UnitTestHelper() {
    }

    /**
     * <p>
     * Gets the value of a private field in the given class. The field has the given name. The value is retrieved from
     * the given instance.
     * </p>
     *
     * @param type the class which the private field belongs to.
     * @param instance the instance which the private field belongs to.
     * @param name the name of the private field to be retrieved.
     *
     * @return the value of the private field or <code>null</code> if any error occurs.
     */
    public static Object getPrivateField(Class type, Object instance, String name) {
        Field field = null;
        Object obj = null;

        try {
            // Get the reflection of the field and get the value
            field = type.getDeclaredField(name);
            field.setAccessible(true);
            obj = field.get(instance);
        } catch (NoSuchFieldException e) {
            // ignore
        } catch (IllegalAccessException e) {
            // ignore
        } finally {
            if (field != null) {
                // Reset the accessibility
                field.setAccessible(false);
            }
        }

        return obj;
    }

    /**
     * <p>
     * Clears the content in database.
     * </p>
     *
     * @param connection the database connection used to access database.
     *
     * @throws SQLException if database error occurs.
     */
    public static void clearSimpleDatabase(Connection connection) throws SQLException {
        Statement statement = connection.createStatement();

        try {
            statement.executeUpdate("DELETE FROM WORD_POSITION;");
            statement.executeUpdate("DELETE FROM WORD;");
            statement.executeUpdate("DELETE FROM DELIMITER;");
            statement.executeUpdate("DELETE FROM DOCUMENT_COLLECTION_DOCUMENT_XREF;");
            statement.executeUpdate("DELETE FROM DOCUMENT_COLLECTION;");
            statement.executeUpdate("DELETE FROM DOCUMENT;");
        } finally {
            statement.close();
        }
    }

    /**
     * <p>
     * Gets the count of records of given table in database.
     * </p>
     *
     * @param connection the database connection used to access database.
     * @param tableName the name of the table.
     *
     * @return the count of records of given table in database.
     *
     * @throws SQLException if database error occurs.
     */
    public static int getRecordCount(Connection connection, String tableName) throws SQLException {
        Statement statement = connection.createStatement();
        ResultSet resultSet = null;

        try {
            resultSet = statement.executeQuery("Select * FROM " + tableName);

            int ret = 0;

            while (resultSet.next()) {
                ret++;
            }

            return ret;
        } finally {
            resultSet.close();
            statement.close();
        }
    }

    /**
     * <p>
     * Clears the content in database.
     * </p>
     *
     * @param connection the database connection used to access database.
     *
     * @throws SQLException if database error occurs.
     */
    public static void clearFastDatabase(Connection connection) throws SQLException {
        Statement statement = connection.createStatement();

        try {
            statement.executeUpdate("DELETE FROM WORD_DATA;");
            statement.executeUpdate("DELETE FROM DELIMITER;");
            statement.executeUpdate("DELETE FROM DOCUMENT_COLLECTION_DOCUMENT_XREF;");
            statement.executeUpdate("DELETE FROM DOCUMENT_COLLECTION;");
            statement.executeUpdate("DELETE FROM DOCUMENT;");
        } finally {
            statement.close();
        }
    }

    /**
     * <p>
     * Clears the configuration used in tests.
     * </p>
     *
     * @throws ConfigManagerException if error occurs when clearing the configuration.
     */
    public static void clearConfiguration() throws ConfigManagerException {
        ConfigManager configManager = ConfigManager.getInstance();

        for (Iterator iter = configManager.getAllNamespaces(); iter.hasNext();) {
            configManager.removeNamespace((String) iter.next());
        }
    }

    /**
     * <p>
     * Adds a valid configuration.
     * </p>
     *
     * @throws ConfigManagerException if error occurs when adding the configuration.
     */
    public static void addValidConfiguration() throws ConfigManagerException {
        ConfigManager configManager = ConfigManager.getInstance();
        configManager.add("DBConnectionFactory.xml");
        configManager.add("ValidXmlPersistenceConfig.xml");
        configManager.add("ValidSimpleAccessDBPersistenceConfig.xml");
        configManager.add("ValidFastAccessDBPersistenceConfig.xml");
    }

    /**
     * <p>
     * Adds a config file into the config manager.
     * </p>
     *
     * @param fileName the name of the config file.
     *
     * @throws ConfigManagerException if error occurs when adding the configuration.
     */
    public static void addConfigurationFile(String fileName) throws ConfigManagerException {
        ConfigManager configManager = ConfigManager.getInstance();
        configManager.add(fileName);
    }

    /**
     * Creates the <code>CollectionIndex</code> instance using the given document index and collection id.
     *
     * @param documentIndexes the given document indexes to create the<code>CollectionIndex</code> instance.
     * @param collectionId the given collection id to create the <code>CollectionIndex</code> instance.
     *
     * @return the created <code>CollectionIndex</code> instance using the given document index and collection id.
     */
    public static CollectionIndex createCollectionIndex(DocumentIndex[] documentIndexes, String collectionId) {
        Map words = new HashMap();
        Set allDocumentIds = new HashSet();

        for (int i = 0; i < documentIndexes.length; i++) {
            for (Iterator iter = documentIndexes[i].getWords().keySet().iterator(); iter.hasNext();) {
                CollationKey key = (CollationKey) iter.next();
                Set documentIds = (Set) words.get(key);

                if (documentIds == null) {
                    // The word does not exist, create a new list
                    documentIds = new HashSet();
                    words.put(key, documentIds);
                }

                // Mark the word's existence in the document
                documentIds.add(documentIndexes[i].getWordSourceId());
            }

            allDocumentIds.add(documentIndexes[i].getWordSourceId());
        }

        // Create the index
        return new CollectionIndex(null, words, allDocumentIds, collectionId);
    }
}
