/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;

import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.UnitTestHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;

import junit.framework.TestCase;

import java.sql.Connection;


/**
 * <p>
 * Tests functionality and error cases of <code>AbstractDBIndexPersistence</code> class. A mock sub-class is used for
 * testing.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class AbstractDBIndexPersistenceUnitTest extends TestCase {
    /** Represents the namespace to create the new <code>SimpleAccessDBIndexPersistence</code> instance for testing. */
    private static final String NAMESPACE =
        "com.topcoder.document.index.persistence.impl.db.SimpleAccessDBIndexPersistence";

    /** Represents the namespace to load DB connection factory configuration. */
    private static final String DB_NAMESPACE = "com.topcoder.document.index.persistence.impl.db";

    /** Represents the connection name to get DB connection factory configuration. */
    private static final String CONN_NAME = "SimpleAccessDBIndexPersistence";

    /** Represents a DB connection factory to create connections. */
    private DBConnectionFactory factory;

    /** Represents the database connection to access database. */
    private Connection connection;

    /** Represents the <code>AbstractDBIndexPersistence</code> instance used in tests. */
    private AbstractDBIndexPersistence persistence;

    /**
     * <p>
     * Sets up the test environment. Valid configurations are loaded. A mock persistence is created.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void setUp() throws Exception {
        UnitTestHelper.clearConfiguration();
        UnitTestHelper.addValidConfiguration();
        persistence = new MockAbstractDBIndexPersistence(NAMESPACE);

        factory = new DBConnectionFactoryImpl(DB_NAMESPACE);
        connection = factory.createConnection(CONN_NAME);

        // add the invalid config file
        UnitTestHelper.addConfigurationFile("InValidDBPersistenceConfig.xml");
    }

    /**
     * Removes the configuration if it exists.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void tearDown() throws Exception {
        UnitTestHelper.clearConfiguration();

        connection.close();
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the given namespace is null,
     * IllegalArgumentException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_NullNamespace() throws Exception {
        try {
            new MockAbstractDBIndexPersistence(null);
            fail("IllegalArgumentException is expected.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the given namespace is empty,
     * IllegalArgumentException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_EmptyNamespace() throws Exception {
        try {
            new MockAbstractDBIndexPersistence(" ");
            fail("IllegalArgumentException is expected.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the given namespace does not exist,
     * PersistenceConfigurationException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_NotExistNamespace() throws Exception {
        try {
            new MockAbstractDBIndexPersistence("NotExist");
            fail("PersistenceConfigurationException is expected.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the property value of
     * ConnectionFactoryClassName is invalid, PersistenceConfigurationException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_NotExistClassNameProperyValue() throws Exception {
        try {
            new MockAbstractDBIndexPersistence("NotExistClassName");
            fail("PersistenceConfigurationException is expected.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the property value of
     * ConnectionFactoryClassName is invalid, PersistenceConfigurationException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_NotCorrectClassNameProperyValue() throws Exception {
        try {
            new MockAbstractDBIndexPersistence("NotCorrectClassName");
            fail("PersistenceConfigurationException is expected.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the property value of
     * ConnectionFactoryNamespace is invalid, PersistenceConfigurationException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_NotExistNamespaceProperyValue() throws Exception {
        try {
            new MockAbstractDBIndexPersistence("NotExistNamespace");
            fail("PersistenceConfigurationException is expected.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the property value of
     * ConnectionFactoryNamespace is invalid, PersistenceConfigurationException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_NotCorrectNamespaceProperyValue() throws Exception {
        try {
            new MockAbstractDBIndexPersistence("NotCorrectNamespace");
            fail("PersistenceConfigurationException is expected.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(String)</code> when the property value of
     * ConnectionFactoryNamespace is missing, PersistenceConfigurationException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_MissingNamespaceProperyValue() throws Exception {
        try {
            new MockAbstractDBIndexPersistence("MissingNamespace");
            fail("PersistenceConfigurationException is expected.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the accuracy of constructor <code>AbstractDBIndexPersistence(String)</code>.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_StringAccuracy() throws Exception {
        assertTrue("The AbstractDBIndexPersistence should implements IndexPersistence.",
            persistence instanceof IndexPersistence);
        assertEquals("The connectionName field should be correct.", CONN_NAME,
            UnitTestHelper.getPrivateField(persistence.getClass().getSuperclass(), persistence, "connectionName"));
        assertNotNull("The connectionFactory field should be set.",
            UnitTestHelper.getPrivateField(persistence.getClass().getSuperclass(), persistence, "connectionFactory"));
    }

    /**
     * <p>
     * Tests the constructor <code>AbstractDBIndexPersistence(DBConnectionFactory, String)</code> when the given
     * connectionFactory is null, IllegalArgumentException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_NullConnectionFactory() throws Exception {
        try {
            new MockAbstractDBIndexPersistence(null, CONN_NAME);
            fail("IllegalArgumentException is expected.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the accuracy of constructor <code>AbstractDBIndexPersistence(DBConnectionFactory, String)</code>.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistence_DBConnectionFactoryStringAccuracy() throws Exception {
        persistence = new MockAbstractDBIndexPersistence(this.factory, CONN_NAME);
        assertTrue("The AbstractDBIndexPersistence should implements IndexPersistence.",
            persistence instanceof IndexPersistence);
        assertEquals("The connectionName field should be correct.", CONN_NAME,
            UnitTestHelper.getPrivateField(persistence.getClass().getSuperclass(), persistence, "connectionName"));
        assertEquals("The connectionFactory field should be set.", this.factory,
            UnitTestHelper.getPrivateField(persistence.getClass().getSuperclass(), persistence, "connectionFactory"));
    }

    /**
     * <p>
     * Tests the accuracy of method <code>getConnection()</code>.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetConnection_Accuracy() throws Exception {
        Connection conn = persistence.getConnection();
        assertNotNull("The connection should be got properly.", conn);
        assertFalse("The value of the auto-commit should be set to false.", conn.getAutoCommit());
        conn.close();
    }

    /**
     * <p>
     * Tests the method <code>beginTransaction(Connection)</code> when the given connection is null,
     * IllegalArgumentException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testBeginTransaction_NullConnection() throws Exception {
        try {
            persistence.beginTransaction(null);
            fail("IllegalArgumentException is expected.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the accuracy of method <code>beginTransaction(Connection)</code>.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testBeginTransaction_Accuracy() throws Exception {
        persistence.beginTransaction(connection);
    }

    /**
     * <p>
     * Tests the method <code>commitTransaction(Connection)</code> when the given connection is null,
     * IllegalArgumentException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testCommitTransaction_NullConnection() throws Exception {
        try {
            persistence.commitTransaction(null);
            fail("IllegalArgumentException is expected.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the accuracy of method <code>commitTransaction(Connection)</code>.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testCommitTransaction_Accuracy() throws Exception {
        connection.setAutoCommit(false);
        persistence.commitTransaction(connection);
    }

    /**
     * <p>
     * Tests the method <code>rollbackTransaction(Connection)</code> when the given connection is null,
     * IllegalArgumentException is expected.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRollbackTransaction_NullConnection() throws Exception {
        try {
            persistence.rollbackTransaction(null);
            fail("IllegalArgumentException is expected.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * <p>
     * Tests the accuracy of method <code>rollbackTransaction(Connection)</code>.
     * </p>
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRollbackTransaction_Accuracy() throws Exception {
        connection.setAutoCommit(false);
        persistence.rollbackTransaction(connection);
    }
}
